import os
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from brainglobe_utils.cells.cells import Cell
from brainglobe_utils.general.system import get_num_processes
from tensorflow import keras

from cellfinder.core import logger, types
from cellfinder.core.classify.cube_generator import CubeGeneratorFromFile
from cellfinder.core.classify.tools import get_model
from cellfinder.core.train.train_yml import depth_type, models


def main(
    points: List[Cell],
    signal_array: types.array,
    background_array: types.array,
    n_free_cpus: int,
    voxel_sizes: Tuple[int, int, int],
    network_voxel_sizes: Tuple[int, int, int],
    batch_size: int,
    cube_height: int,
    cube_width: int,
    cube_depth: int,
    trained_model: Optional[os.PathLike],
    model_weights: Optional[os.PathLike],
    network_depth: depth_type,
    max_workers: int = 3,
    *,
    callback: Optional[Callable[[int], None]] = None,
) -> List:
    """
    Parameters
    ----------
    callback : Callable[int], optional
        A callback function that is called during classification. Called with
        the batch number once that batch has been classified.
    """
    if signal_array.ndim != 3:
        raise IOError("Signal data must be 3D")
    if background_array.ndim != 3:
        raise IOError("Background data must be 3D")

    if callback is not None:
        callbacks = [BatchEndCallback(callback)]
    else:
        callbacks = None

    # Too many workers doesn't increase speed, and uses huge amounts of RAM
    workers = get_num_processes(
        min_free_cpu_cores=n_free_cpus, n_max_processes=max_workers
    )

    logger.debug("Initialising cube generator")
    inference_generator = CubeGeneratorFromFile(
        points,
        signal_array,
        background_array,
        voxel_sizes,
        network_voxel_sizes,
        batch_size=batch_size,
        cube_width=cube_width,
        cube_height=cube_height,
        cube_depth=cube_depth,
    )

    model = get_model(
        existing_model=trained_model,
        model_weights=model_weights,
        network_depth=models[network_depth],
        inference=True,
    )

    logger.info("Running inference")
    predictions = model.predict(
        inference_generator,
        use_multiprocessing=True,
        workers=workers,
        verbose=True,
        callbacks=callbacks,
    )
    predictions = predictions.round()
    predictions = predictions.astype("uint16")

    predictions = np.argmax(predictions, axis=1)
    points_list = []

    # only go through the "extractable" points
    for idx, cell in enumerate(inference_generator.ordered_points):
        cell.type = predictions[idx] + 1
        points_list.append(cell)

    return points_list


class BatchEndCallback(keras.callbacks.Callback):
    def __init__(self, callback: Callable[[int], None]):
        self._callback = callback

    def on_predict_batch_end(
        self, batch: int, logs: Optional[Dict[str, Any]] = None
    ) -> None:
        self._callback(batch)



from napari.qt.threading import WorkerBase, WorkerBaseSignals
from qtpy.QtCore import Signal

from cellfinder.core.main import main as cellfinder_run

from .detect_containers import (
    ClassificationInputs,
    DataInputs,
    DetectionInputs,
    MiscInputs,
)


class MyWorkerSignals(WorkerBaseSignals):
    """
    Signals used by the Worker class below.
    """

    # Emits (label, max, value) for the progress bar
    update_progress_bar = Signal(str, int, int)


class Worker(WorkerBase):
    """
    Runs cellfinder in a separate thread, to prevent GUI blocking.

    Also handles callbacks between the worker thread and main napari GUI thread
    to update a progress bar.
    """

    def __init__(
        self,
        data_inputs: DataInputs,
        detection_inputs: DetectionInputs,
        classification_inputs: ClassificationInputs,
        misc_inputs: MiscInputs,
    ):
        super().__init__(SignalsClass=MyWorkerSignals)
        self.data_inputs = data_inputs
        self.detection_inputs = detection_inputs
        self.classification_inputs = classification_inputs
        self.misc_inputs = misc_inputs

    def work(self) -> list:
        self.update_progress_bar.emit("Setting up detection...", 1, 0)

        def detect_callback(plane: int) -> None:
            self.update_progress_bar.emit(
                "Detecting cells",
                self.data_inputs.nplanes,
                plane + 1,
            )

        def detect_finished_callback(points: list) -> None:
            self.npoints_detected = len(points)
            self.update_progress_bar.emit("Setting up classification...", 1, 0)

        def classify_callback(batch: int) -> None:
            self.update_progress_bar.emit(
                "Classifying cells",
                # Default cellfinder-core batch size is 32. This seems to give
                # a slight underestimate of the number of batches though, so
                # allow for batch number to go over this
                max(self.npoints_detected // 32 + 1, batch + 1),
                batch + 1,
            )

        result = cellfinder_run(
            **self.data_inputs.as_core_arguments(),
            **self.detection_inputs.as_core_arguments(),
            **self.classification_inputs.as_core_arguments(),
            **self.misc_inputs.as_core_arguments(),
            detect_callback=detect_callback,
            classify_callback=classify_callback,
            detect_finished_callback=detect_finished_callback,
        )
        self.update_progress_bar.emit("Finished classification", 1, 1)
        return result



from pathlib import Path
from typing import List, Optional, Tuple

import napari
import numpy as np
import tifffile
from brainglobe_napari_io.cellfinder.utils import convert_layer_to_cells
from brainglobe_utils.cells.cells import Cell
from brainglobe_utils.general.system import delete_directory_contents
from brainglobe_utils.IO.yaml import save_yaml
from brainglobe_utils.qtpy.dialog import display_warning
from brainglobe_utils.qtpy.interaction import add_button, add_combobox
from magicgui.widgets import ProgressBar
from napari.qt.threading import thread_worker
from napari.utils.notifications import show_info
from qtpy import QtCore
from qtpy.QtWidgets import (
    QComboBox,
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QLabel,
    QWidget,
)

# Constants used throughout
WINDOW_HEIGHT = 750
WINDOW_WIDTH = 1500
COLUMN_WIDTH = 150


class CurationWidget(QWidget):
    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        cube_depth: int = 20,
        cube_width: int = 50,
        cube_height: int = 50,
        network_voxel_sizes: Tuple[int, int, int] = (5, 1, 1),
        n_free_cpus: int = 2,
        save_empty_cubes: bool = False,
        max_ram=None,
    ):
        super(CurationWidget, self).__init__()

        self.non_cells_to_extract = None
        self.cells_to_extract = None

        self.cube_depth = cube_depth
        self.cube_width = cube_width
        self.cube_height = cube_height
        self.network_voxel_sizes = network_voxel_sizes
        self.n_free_cpus = n_free_cpus
        self.save_empty_cubes = save_empty_cubes
        self.max_ram = max_ram
        self.voxel_sizes = [5, 2, 2]
        self.batch_size = 32
        self.viewer = viewer

        self.signal_layer = None
        self.background_layer = None
        self.training_data_cell_layer = None
        self.training_data_non_cell_layer = None

        self.image_layer_names = self._get_layer_names()
        self.point_layer_names = self._get_layer_names(
            layer_type=napari.layers.Points
        )

        self.output_directory: Optional[Path] = None

        self.setup_main_layout()
        self.setup_keybindings()

        @self.viewer.layers.events.connect
        def update_layer_list(v: napari.viewer.Viewer):
            """
            Update internal list of layers whenever the napari layers list
            is updated.
            """
            self.image_layer_names = self._get_layer_names()
            self.point_layer_names = self._get_layer_names(
                layer_type=napari.layers.Points
            )
            self._update_combobox_options(
                self.signal_image_choice, self.image_layer_names
            )
            self._update_combobox_options(
                self.background_image_choice, self.image_layer_names
            )
            self._update_combobox_options(
                self.training_data_cell_choice, self.point_layer_names
            )
            self._update_combobox_options(
                self.training_data_non_cell_choice, self.point_layer_names
            )

    @staticmethod
    def _update_combobox_options(combobox: QComboBox, options_list: List[str]):
        original_text = combobox.currentText()
        combobox.clear()
        combobox.addItems(options_list)
        combobox.setCurrentText(original_text)

    def _get_layer_names(
        self,
        layer_type: napari.layers.Layer = napari.layers.Image,
        default: str = "",
    ) -> List[str]:
        """
        Get list of layer names of a given layer type.
        """
        layer_names = [
            layer.name
            for layer in self.viewer.layers
            if type(layer) == layer_type
        ]

        if layer_names:
            return [default] + layer_names
        else:
            return [default]

    def setup_main_layout(self):
        """
        Construct main layout of widget.
        """
        self.layout = QGridLayout()
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        self.layout.setSpacing(4)

        self.add_loading_panel(1)

        self.status_label = QLabel()
        row, col = 7, 0
        self.layout.addWidget(self.status_label, row, col)
        self.update_status_label("Ready")

        self.progress_bar = ProgressBar()
        row, col = 8, 0
        self.layout.addWidget(self.progress_bar.native, row, col)

        self.setLayout(self.layout)

    def add_loading_panel(self, row: int, column: int = 0):
        self.load_data_panel = QGroupBox("Load data")
        self.load_data_layout = QGridLayout()
        self.load_data_layout.setSpacing(15)
        self.load_data_layout.setContentsMargins(10, 10, 10, 10)
        self.load_data_layout.setAlignment(QtCore.Qt.AlignBottom)

        self.signal_image_choice, _ = add_combobox(
            self.load_data_layout,
            "Signal image",
            self.image_layer_names,
            1,
            callback=self.set_signal_image,
        )
        self.background_image_choice, _ = add_combobox(
            self.load_data_layout,
            "Background image",
            self.image_layer_names,
            2,
            callback=self.set_background_image,
        )
        self.training_data_cell_choice, _ = add_combobox(
            self.load_data_layout,
            "Training data (cells)",
            self.point_layer_names,
            3,
            callback=self.set_training_data_cell,
        )
        self.training_data_non_cell_choice, _ = add_combobox(
            self.load_data_layout,
            "Training_data (non_cells)",
            self.point_layer_names,
            row=4,
            callback=self.set_training_data_non_cell,
        )
        self.mark_as_cell_button = add_button(
            "Mark as cell(s)",
            self.load_data_layout,
            self.mark_as_cell,
            row=5,
            tooltip="Mark all selected points as non cell. Shortcut: 'c'",
        )
        self.mark_as_non_cell_button = add_button(
            "Mark as non cell(s)",
            self.load_data_layout,
            self.mark_as_non_cell,
            row=5,
            column=1,
            tooltip="Mark all selected points as non cell. Shortcut: 'x'",
        )
        self.add_training_data_button = add_button(
            "Add training data layers",
            self.load_data_layout,
            self.add_training_data,
            row=6,
        )
        self.save_training_data_button = add_button(
            "Save training data",
            self.load_data_layout,
            self.save_training_data,
            row=6,
            column=1,
        )
        self.load_data_layout.setColumnMinimumWidth(0, COLUMN_WIDTH)
        self.load_data_panel.setLayout(self.load_data_layout)
        self.load_data_panel.setVisible(True)
        self.layout.addWidget(self.load_data_panel, row, column, 1, 1)

    def setup_keybindings(self):
        self.viewer.bind_key("c", self.mark_as_cell)
        self.viewer.bind_key("x", self.mark_as_non_cell)

    def set_signal_image(self):
        """
        Set signal layer from current signal text box selection.
        """
        if self.signal_image_choice.currentText() != "":
            self.signal_layer = self.viewer.layers[
                self.signal_image_choice.currentText()
            ]

    def set_background_image(self):
        """
        Set background layer from current background text box selection.
        """
        if self.background_image_choice.currentText() != "":
            self.background_layer = self.viewer.layers[
                self.background_image_choice.currentText()
            ]

    def set_training_data_cell(self):
        """
        Set cell training data from current training data text box selection.
        """
        if self.training_data_cell_choice.currentText() != "":
            self.training_data_cell_layer = self.viewer.layers[
                self.training_data_cell_choice.currentText()
            ]
            self.training_data_cell_layer.metadata["point_type"] = Cell.CELL
            self.training_data_cell_layer.metadata["training_data"] = True

    def set_training_data_non_cell(self):
        """
        Set non-cell training data from current training data text box
        selection.
        """
        if self.training_data_non_cell_choice.currentText() != "":
            self.training_data_non_cell_layer = self.viewer.layers[
                self.training_data_non_cell_choice.currentText()
            ]
            self.training_data_non_cell_layer.metadata["point_type"] = (
                Cell.UNKNOWN
            )
            self.training_data_non_cell_layer.metadata["training_data"] = True

    def add_training_data(self):
        cell_name = "Training data (cells)"
        non_cell_name = "Training data (non cells)"

        overwrite = False
        if self.training_data_cell_layer or self.training_data_non_cell_layer:
            overwrite = display_warning(
                self,
                "Training data layers exist",
                "Training data layers already exist,  "
                "overwrite with empty layers?.",
            )
        else:
            if self.training_data_cell_layer:
                self.training_data_cell_layer.remove()
            self._add_training_data_layers(cell_name, non_cell_name)

        if overwrite:
            try:
                self.viewer.layers.remove(cell_name)
                self.viewer.layers.remove(non_cell_name)
            except ValueError:
                pass

            self._add_training_data_layers(cell_name, non_cell_name)

    def _add_training_data_layers(self, cell_name: str, non_cell_name: str):
        self.training_data_cell_layer = self.viewer.add_points(
            None,
            ndim=3,
            symbol="ring",
            n_dimensional=True,
            size=15,
            opacity=0.6,
            face_color="lightgoldenrodyellow",
            name=cell_name,
            metadata=dict(point_type=Cell.CELL, training_data=True),
        )
        self.training_data_cell_choice.setCurrentText(cell_name)

        self.training_data_non_cell_layer = self.viewer.add_points(
            None,
            ndim=3,
            symbol="ring",
            n_dimensional=True,
            size=15,
            opacity=0.6,
            face_color="lightskyblue",
            name=non_cell_name,
            metadata=dict(point_type=Cell.UNKNOWN, training_data=True),
        )
        self.training_data_non_cell_choice.setCurrentText(non_cell_name)

    def mark_as_cell(self, viewer=None):
        self.mark_point_as_type("cell")

    def mark_as_non_cell(self, viewer=None):
        self.mark_point_as_type("non-cell")

    def mark_point_as_type(self, point_type: str):
        if not (
            self.training_data_cell_layer and self.training_data_non_cell_layer
        ):
            show_info(
                "No training data layers have been chosen. "
                "Please add training data layers. ",
            )
            return

        if len(self.viewer.layers.selection) == 1:
            layer = list(self.viewer.layers.selection)[0]
            if type(layer) == napari.layers.Points:
                if len(layer.data) > 0:
                    if point_type == "cell":
                        destination_layer = self.training_data_cell_layer
                    else:
                        destination_layer = self.training_data_non_cell_layer
                    show_info(
                        f"Adding {len(layer.selected_data)} "
                        f"points to layer: {destination_layer.name}"
                    )

                    for selected_point in layer.selected_data:
                        destination_layer.data = np.vstack(
                            (
                                destination_layer.data,
                                layer.data[selected_point],
                            )
                        )

                else:
                    show_info(
                        "No points are selected in the current layer. "
                        "Please select some points.",
                    )

            else:
                show_info(
                    "This is not a points layer. "
                    "Please choose a points layer, and select some points.",
                )
        elif len(self.viewer.layers.selected) == 0:
            show_info(
                "No layers are selected. "
                "Please choose a single points layer, and select some points.",
            )
        else:
            show_info(
                "More than one layer is selected. "
                "Please choose a single points layer, and select some points.",
            )

    def save_training_data(
        self,
        *,
        block: bool = False,
        prompt_for_directory: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        block :
            If `True` block execution until all cubes are saved.
        prompt_for_directory :
            If `True` show a file dialog for the user to select a directory.
        """

        if self.is_data_extractable():
            if prompt_for_directory:
                self.get_output_directory()
                # if the directory is not empty
                if any(self.output_directory.iterdir()):
                    choice = display_warning(
                        self,
                        "About to save training data",
                        "Existing files will be will be deleted. Proceed?",
                    )
                    if not choice:
                        return
            if self.output_directory is not None:
                self.__prep_directories_for_save()
                self.__extract_cubes(block=block)
                self.__save_yaml_file()
                show_info("Done")

            self.update_status_label("Ready")

    def __prep_directories_for_save(self):
        self.yaml_filename = self.output_directory / "training.yml"
        self.cell_cube_dir = self.output_directory / "cells"
        self.no_cell_cube_dir = self.output_directory / "non_cells"

        self.__delete_existing_saved_training_data()

    def __delete_existing_saved_training_data(self):
        self.yaml_filename.unlink(missing_ok=True)
        for directory in (
            self.cell_cube_dir,
            self.no_cell_cube_dir,
        ):
            if directory.exists():
                delete_directory_contents(directory)
            else:
                directory.mkdir(exist_ok=True, parents=True)

    def __extract_cubes(self, *, block=False):
        """
        Parameters
        ----------
        block :
            If `True` block execution until all cubes are saved.
        """
        self.update_status_label("Extracting cubes")
        self.convert_layers_to_cells()

        if block:
            cubes = self.extract_cubes()
            while True:
                try:
                    next(cubes)
                except StopIteration:
                    break
        else:

            @thread_worker(connect={"yielded": self.update_progress})
            def extract_cubes():
                yield from self.extract_cubes()

            extract_cubes()

    def is_data_extractable(self) -> bool:
        if (
            self.check_training_data_exists()
            and self.check_image_data_for_extraction()
        ):
            return True
        else:
            return False

    def check_image_data_for_extraction(self) -> bool:
        if self.signal_layer and self.background_layer:
            if (
                self.signal_layer.data.shape
                == self.background_layer.data.shape
            ):
                return True
            else:
                show_info(
                    "Please ensure both signal and background images are the "
                    "same size and shape.",
                )
                return False

        else:
            show_info(
                "Please ensure both signal and background images are loaded "
                "into napari, and selected in the sidebar. ",
            )
            return False

    def check_training_data_exists(self) -> bool:
        if not (
            self.training_data_cell_layer or self.training_data_non_cell_layer
        ):
            show_info(
                "No training data layers have been added. "
                "Please add a layer and annotate some points.",
            )
            return False
        else:
            if (
                len(self.training_data_cell_layer.data) > 0
                or len(self.training_data_non_cell_layer.data) > 0
            ):
                return True
            else:
                show_info(
                    "No training data points have been added. "
                    "Please annotate some points.",
                )
                return False

    def get_output_directory(self):
        """
        Shows file dialog to choose output directory
        """
        self.update_status_label("Setting output directory...")
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        self.output_directory = QFileDialog.getExistingDirectory(
            self,
            "Select output directory",
            options=options,
        )
        if self.output_directory != "":
            self.output_directory = Path(self.output_directory)
        else:
            self.output_directory = None

    def convert_layers_to_cells(self):
        self.cells_to_extract = convert_layer_to_cells(
            self.training_data_cell_layer.data
        )
        self.non_cells_to_extract = convert_layer_to_cells(
            self.training_data_non_cell_layer.data, cells=False
        )

        self.cells_to_extract = list(set(self.cells_to_extract))
        self.non_cells_to_extract = list(set(self.non_cells_to_extract))

    def __save_yaml_file(self):
        yaml_section = [
            {
                "cube_dir": str(self.cell_cube_dir),
                "cell_def": "",
                "type": "cell",
                "signal_channel": 0,
                "bg_channel": 1,
            },
            {
                "cube_dir": str(self.no_cell_cube_dir),
                "cell_def": "",
                "type": "no_cell",
                "signal_channel": 0,
                "bg_channel": 1,
            },
        ]

        yaml_contents = {"data": yaml_section}
        save_yaml(yaml_contents, self.yaml_filename)

    def update_progress(self, attributes: dict):
        """
        Update progress bar with ``attributes``.
        """
        for attr in attributes:
            self.progress_bar.__setattr__(attr, attributes[attr])

    def update_status_label(self, label: str):
        self.status_label.setText(label)

    def extract_cubes(self):
        """
        Yields
        ------
        dict
            Attributes used to update a progress bar. The keys can be any of
            the properties of `magicgui.widgets.ProgressBar`.
        """
        from cellfinder.core.classify.cube_generator import (
            CubeGeneratorFromFile,
        )

        to_extract = {
            "cells": self.cells_to_extract,
            "non_cells": self.non_cells_to_extract,
        }

        directories = {
            "cells": self.cell_cube_dir,
            "non_cells": self.no_cell_cube_dir,
        }

        for cell_type in ["cells", "non_cells"]:
            cell_type_output_directory = directories[cell_type]
            cell_list = to_extract[cell_type]

            self.update_status_label(f"Saving {cell_type}...")

            cube_generator = CubeGeneratorFromFile(
                cell_list,
                self.signal_layer.data,
                self.background_layer.data,
                self.voxel_sizes,
                self.network_voxel_sizes,
                batch_size=self.batch_size,
                cube_width=self.cube_width,
                cube_height=self.cube_height,
                cube_depth=self.cube_depth,
                extract=True,
            )
            # Set up progress bar
            yield {
                "value": 0,
                "min": 0,
                "max": len(cube_generator),
            }

            for i, (image_batch, batch_info) in enumerate(cube_generator):
                image_batch = image_batch.astype(np.int16)

                for point, point_info in zip(image_batch, batch_info):
                    point = np.moveaxis(point, 2, 0)
                    for channel in range(point.shape[-1]):
                        save_cube(
                            point,
                            point_info,
                            channel,
                            cell_type_output_directory,
                        )

                # Update progress bar
                yield {"value": i + 1}

            self.update_status_label("Finished saving cubes")


def save_cube(
    array: np.ndarray, point_info: dict, channel: int, output_directory: Path
):
    filename = (
        f"pCellz{point_info['z']}y{point_info['y']}"
        f"x{point_info['x']}Ch{channel}.tif"
    )
    tifffile.imwrite(output_directory / filename, array[:, :, :, channel])



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from magicgui.types import FileDialogMode

from cellfinder.core.download.models import model_weight_urls
from cellfinder.core.train.train_yml import models
from cellfinder.napari.input_container import InputContainer
from cellfinder.napari.utils import html_label_widget


@dataclass
class TrainingDataInputs(InputContainer):
    """Container for Training Data input widgets"""

    yaml_files: Path = Path.home()
    output_directory: Path = Path.home()

    def as_core_arguments(self) -> dict:
        arguments = super().as_core_arguments()
        arguments["output_dir"] = arguments.pop("output_directory")
        arguments["yaml_file"] = arguments.pop("yaml_files")
        return arguments

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            data_options=html_label_widget("Training Data:"),
            yaml_files=cls._custom_widget(
                "yaml_files",
                custom_label="YAML files",
                mode=FileDialogMode.EXISTING_FILES,
                filter="*.yml",
            ),
            output_directory=cls._custom_widget(
                "output_directory", mode=FileDialogMode.EXISTING_DIRECTORY
            ),
        )


@dataclass
class OptionalNetworkInputs(InputContainer):
    """Container for Optional Network input widgets"""

    trained_model: Optional[Path] = Path.home()
    model_weights: Optional[Path] = Path.home()
    model_depth: str = list(models.keys())[2]
    pretrained_model: str = str(list(model_weight_urls.keys())[0])

    def as_core_arguments(self) -> dict:
        arguments = super().as_core_arguments()
        arguments["model"] = arguments.pop("pretrained_model")
        arguments["network_depth"] = arguments.pop("model_depth")
        return arguments

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            network_options=html_label_widget("Network (optional)"),
            trained_model=cls._custom_widget("trained_model"),
            model_weights=cls._custom_widget("model_weights"),
            model_depth=cls._custom_widget(
                "model_depth", choices=list(models.keys())
            ),
            pretrained_model=cls._custom_widget(
                "pretrained_model",
                choices=list(model_weight_urls.keys()),
            ),
        )


@dataclass
class OptionalTrainingInputs(InputContainer):
    continue_training: bool = False
    augment: bool = True
    tensorboard: bool = False
    save_weights: bool = False
    save_checkpoints: bool = True
    save_progress: bool = True
    epochs: int = 100
    learning_rate: float = 1e-4
    batch_size: int = 16
    test_fraction: float = 0.1

    def as_core_arguments(self) -> dict:
        arguments = super().as_core_arguments()
        arguments["no_augment"] = not arguments.pop("augment")
        arguments["no_save_checkpoints"] = not arguments.pop(
            "save_checkpoints"
        )
        return arguments

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            training_options=html_label_widget("Training (optional)"),
            continue_training=cls._custom_widget("continue_training"),
            augment=cls._custom_widget("augment"),
            tensorboard=cls._custom_widget("tensorboard"),
            save_weights=cls._custom_widget("save_weights"),
            save_checkpoints=cls._custom_widget("save_checkpoints"),
            save_progress=cls._custom_widget("save_progress"),
            epochs=cls._custom_widget("epochs"),
            learning_rate=cls._custom_widget("learning_rate", step=1e-4),
            batch_size=cls._custom_widget("batch_size"),
            test_fraction=cls._custom_widget(
                "test_fraction", step=0.05, min=0.05, max=0.95
            ),
        )


@dataclass
class MiscTrainingInputs(InputContainer):
    number_of_free_cpus: int = 2

    def as_core_arguments(self) -> dict:
        return dict(n_free_cpus=self.number_of_free_cpus)

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            misc_options=html_label_widget("Misc (optional):"),
            number_of_free_cpus=cls._custom_widget(
                "number_of_free_cpus", custom_label="Number of free CPUs"
            ),
        )



import math
from typing import Generator, Tuple

import numpy as np
from numba import njit


class TileWalker:
    """
    A class to segment a 2D image into tiles, and mark each of the
    tiles as bright or dark depending on whether the average image
    value in each tile is above a threshold.

    The threshold is set using the tile of data containing the corner (0, 0).
    The mean and standard deviation of this tile is calculated, and
    the threshold set at 1 + mean + (2 * stddev).

    Attributes
    ----------
    bright_tiles_mask :
        An boolean array whose entries correspond to whether each tile is
        bright (1) or dark (0). The values are set in
        self.mark_bright_tiles().
    """

    def __init__(self, img: np.ndarray, soma_diameter: int) -> None:
        self.img = img
        self.img_width, self.img_height = img.shape
        self.tile_width = soma_diameter * 2
        self.tile_height = soma_diameter * 2

        n_tiles_width = math.ceil(self.img_width / self.tile_width)
        n_tiles_height = math.ceil(self.img_height / self.tile_height)
        self.bright_tiles_mask = np.zeros(
            (n_tiles_width, n_tiles_height), dtype=bool
        )

        corner_tile = img[0 : self.tile_width, 0 : self.tile_height]
        corner_intensity = np.mean(corner_tile)
        corner_sd = np.std(corner_tile)
        # add 1 to ensure not 0, as disables
        self.out_of_brain_threshold = (corner_intensity + (2 * corner_sd)) + 1

    def _get_tiles(self) -> Generator[Tuple[int, int, np.ndarray], None, None]:
        """
        Generator that yields tiles of the 2D image.

        Notes
        -----
        The final tile in each dimension can have a smaller size than the
        rest of the tiles if the tile shape does not exactly divide the
        image shape.
        """
        for y in range(
            0, self.img_height - self.tile_height, self.tile_height
        ):
            for x in range(
                0, self.img_width - self.tile_width, self.tile_width
            ):
                tile = self.img[
                    x : x + self.tile_width, y : y + self.tile_height
                ]
                yield x, y, tile

    def mark_bright_tiles(self) -> None:
        """
        Loop through tiles, and if the average value of a tile is
        greater than the intensity threshold mark the tile as bright
        in self.bright_tiles_mask.
        """
        threshold = self.out_of_brain_threshold
        if threshold == 0:
            return

        for x, y, tile in self._get_tiles():
            if not is_low_average(tile, threshold):
                mask_x = x // self.tile_width
                mask_y = y // self.tile_height
                self.bright_tiles_mask[mask_x, mask_y] = True


@njit
def is_low_average(tile: np.ndarray, threshold: float) -> bool:
    """
    Return `True` if the average value of *tile* is below *threshold*.
    """
    avg = np.mean(tile)
    return avg < threshold



import argparse

BRAINGLOBE_WORKFLOWS = "https://github.com/brainglobe/brainglobe-workflows"
NEW_NAME = "brainmapper"
BLOG_POST = "https://brainglobe.info/blog/version1/core_and_napari_merge.html"


def cli_catch() -> None:
    """
    Will be executed if the user attempts to run 'cellfinder' from the
    command-line.
    Reports that this CLI tool is now only available as a workflow from
    brainglobe-workflows, and has a new name to boot.
    Relevant links are provided.

    Argparse is used so that the CLI arguments are caught, but they are
    not processed.
    The --help feature is overwritten to force argparse to display the
    migration message regardless of user input.
    """
    parser = argparse.ArgumentParser(
        "cellfinder",
        description=(
            "Migration warning if users are trying to use the cellfinder"
            " name for the workflow/ CLI tool now provided by"
            " brainglobe-workflows."
        ),
        add_help=False,
    )
    parser.add_argument(
        "catch-all",
        nargs="*",
        help=(
            "Catch any input arguments to prevent error throws."
            " All arguments are ignored."
        ),
    )

    print(
        "Hey, it looks like you're trying to run the old command-line tool.",
        "This workflow has been renamed and moved -",
        " you can now find it in the brainglobe-workflows package:\n",
        f"\t{BRAINGLOBE_WORKFLOWS}\n",
        f"\tusing the name {NEW_NAME}\n",
        f"For more information, see our blog post: {BLOG_POST}",
    )
    return



from pathlib import Path


def source_config_cellfinder():
    return Path(__file__).parent.parent / "config" / "cellfinder.conf"


def source_custom_config_cellfinder():
    return Path(__file__).parent.parent / "config" / "cellfinder.conf.custom"



from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Optional


@dataclass
class InputContainer:
    """Base for classes that contain inputs

    Intended to be derived to group specific related widget inputs (e.g from
    the same widget section) into a container. Derived classes should be
    Python data classes.

    Enforces common interfaces for
    - how to get default values for the inputs
    - how inputs are passed to cellfinder core
    - how the inputs are shown in the widget
    """

    @classmethod
    def defaults(cls) -> dict:
        """Returns default values of this class's fields as a dict."""
        # Derived classes are not expected to be particularly
        # slow to instantiate, so use the default constructor
        # to avoid code repetition.
        return asdict(cls())

    @abstractmethod
    def as_core_arguments(self) -> dict:
        """Determines how dataclass fields are passed to cellfinder-core.

        The implementation provided here can be re-used in derived classes, if
        convenient.
        """
        # note that asdict returns a new instance of a dict,
        # so any subsequent modifications of this dict won't affect the class
        # instance
        return asdict(self)

    @classmethod
    def _custom_widget(
        cls, key: str, custom_label: Optional[str] = None, **kwargs
    ) -> dict:
        """
        Represents a field, given by key, as a formatted widget with the
        field's default value.

        The widget label is the capitalized key by default, with underscores
        replaced by spaces, unless custom_label is specified. Keyword
        arguments like step, min, max, ... are passed to napari underneath.
        """
        label = (
            key.replace("_", " ").capitalize()
            if custom_label is None
            else custom_label
        )
        value = cls.defaults()[key]
        return dict(value=value, label=label, **kwargs)

    @classmethod
    @abstractmethod
    def widget_representation(cls) -> dict:
        """What the class will look like as a napari widget"""
        pass



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy

from cellfinder.napari.input_container import InputContainer
from cellfinder.napari.utils import html_label_widget


@dataclass
class DataInputs(InputContainer):
    """Container for image-related ("Data") inputs."""

    signal_array: numpy.ndarray = None
    background_array: numpy.ndarray = None
    voxel_size_z: float = 5
    voxel_size_y: float = 2
    voxel_size_x: float = 2

    def as_core_arguments(self) -> dict:
        """
        Passes voxel size data as one tuple instead of 3 individual floats
        """
        data_input_dict = super().as_core_arguments()
        data_input_dict["voxel_sizes"] = (
            self.voxel_size_z,
            self.voxel_size_y,
            self.voxel_size_x,
        )
        # del operator doesn't affect self, because asdict creates a copy of
        # fields.
        del data_input_dict["voxel_size_z"]
        del data_input_dict["voxel_size_y"]
        del data_input_dict["voxel_size_x"]
        return data_input_dict

    @property
    def nplanes(self):
        return len(self.signal_array)

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            data_options=html_label_widget("Data:"),
            voxel_size_z=cls._custom_widget(
                "voxel_size_z", custom_label="Voxel size (z)"
            ),
            voxel_size_y=cls._custom_widget(
                "voxel_size_y", custom_label="Voxel size (y)"
            ),
            voxel_size_x=cls._custom_widget(
                "voxel_size_x", custom_label="Voxel size (x)"
            ),
        )


@dataclass
class DetectionInputs(InputContainer):
    """Container for cell candidate detection inputs."""

    soma_diameter: float = 16.0
    ball_xy_size: float = 6
    ball_z_size: float = 15
    ball_overlap_fraction: float = 0.6
    log_sigma_size: float = 0.2
    n_sds_above_mean_thresh: int = 10
    soma_spread_factor: float = 1.4
    max_cluster_size: int = 100000

    def as_core_arguments(self) -> dict:
        return super().as_core_arguments()

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            detection_options=html_label_widget("Detection:"),
            soma_diameter=cls._custom_widget("soma_diameter"),
            ball_xy_size=cls._custom_widget(
                "ball_xy_size", custom_label="Ball filter (xy)"
            ),
            ball_z_size=cls._custom_widget(
                "ball_z_size", custom_label="Ball filter (z)"
            ),
            ball_overlap_fraction=cls._custom_widget(
                "ball_overlap_fraction", custom_label="Ball overlap"
            ),
            log_sigma_size=cls._custom_widget(
                "log_sigma_size", custom_label="Filter width"
            ),
            n_sds_above_mean_thresh=cls._custom_widget(
                "n_sds_above_mean_thresh", custom_label="Threshold"
            ),
            soma_spread_factor=cls._custom_widget(
                "soma_spread_factor", custom_label="Cell spread"
            ),
            max_cluster_size=cls._custom_widget(
                "max_cluster_size",
                custom_label="Max cluster",
                min=0,
                max=10000000,
            ),
        )


@dataclass
class ClassificationInputs(InputContainer):
    """Container for classification inputs."""

    use_pre_trained_weights: bool = True
    trained_model: Optional[Path] = Path.home()

    def as_core_arguments(self) -> dict:
        args = super().as_core_arguments()
        del args["use_pre_trained_weights"]
        return args

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            classification_options=html_label_widget("Classification:"),
            use_pre_trained_weights=dict(
                value=cls.defaults()["use_pre_trained_weights"]
            ),
            trained_model=dict(value=cls.defaults()["trained_model"]),
        )


@dataclass
class MiscInputs(InputContainer):
    """Container for miscellaneous inputs."""

    start_plane: int = 0
    end_plane: int = 0
    n_free_cpus: int = 2
    analyse_local: bool = False
    debug: bool = False

    def as_core_arguments(self) -> dict:
        misc_input_dict = super().as_core_arguments()
        del misc_input_dict["debug"]
        del misc_input_dict["analyse_local"]
        return misc_input_dict

    @classmethod
    def widget_representation(cls) -> dict:
        return dict(
            misc_options=html_label_widget("Miscellaneous:"),
            start_plane=cls._custom_widget("start_plane", min=0, max=100000),
            end_plane=cls._custom_widget("end_plane", min=0, max=100000),
            n_free_cpus=cls._custom_widget(
                "n_free_cpus", custom_label="Number of free CPUs"
            ),
            analyse_local=dict(value=cls.defaults()["analyse_local"]),
            debug=dict(value=cls.defaults()["debug"]),
        )



from math import ceil
from pathlib import Path
from typing import Optional

import napari
from magicgui import magicgui
from magicgui.widgets import FunctionGui, ProgressBar
from napari.utils.notifications import show_info
from qtpy.QtWidgets import QScrollArea

from cellfinder.core.classify.cube_generator import get_cube_depth_min_max
from cellfinder.napari.utils import (
    add_layers,
    header_label_widget,
    html_label_widget,
    widget_header,
)

from .detect_containers import (
    ClassificationInputs,
    DataInputs,
    DetectionInputs,
    MiscInputs,
)
from .thread_worker import Worker

NETWORK_VOXEL_SIZES = [5, 1, 1]
CUBE_WIDTH = 50
CUBE_HEIGHT = 20
CUBE_DEPTH = 20

# If using ROI, how many extra planes to analyse
MIN_PLANES_ANALYSE = 0


def detect_widget() -> FunctionGui:
    """
    Create a detection plugin GUI.
    """
    progress_bar = ProgressBar()

    @magicgui(
        header=header_label_widget,
        detection_label=html_label_widget("Cell detection", tag="h3"),
        **DataInputs.widget_representation(),
        **DetectionInputs.widget_representation(),
        **ClassificationInputs.widget_representation(),
        **MiscInputs.widget_representation(),
        call_button=True,
        persist=True,
        reset_button=dict(widget_type="PushButton", text="Reset defaults"),
        scrollable=True,
    )
    def widget(
        header,
        detection_label,
        data_options,
        viewer: napari.Viewer,
        signal_image: napari.layers.Image,
        background_image: napari.layers.Image,
        voxel_size_z: float,
        voxel_size_y: float,
        voxel_size_x: float,
        detection_options,
        soma_diameter: float,
        ball_xy_size: float,
        ball_z_size: float,
        ball_overlap_fraction: float,
        log_sigma_size: float,
        n_sds_above_mean_thresh: int,
        soma_spread_factor: float,
        max_cluster_size: int,
        classification_options,
        trained_model: Optional[Path],
        use_pre_trained_weights: bool,
        misc_options,
        start_plane: int,
        end_plane: int,
        n_free_cpus: int,
        analyse_local: bool,
        debug: bool,
        reset_button,
    ) -> None:
        """
        Run detection and classification.

        Parameters
        ----------
        signal_image : napari.layers.Image
             Image layer containing the labelled cells
        background_image : napari.layers.Image
             Image layer without labelled cells
        voxel_size_z : float
            Size of your voxels in the axial dimension
        voxel_size_y : float
            Size of your voxels in the y direction (top to bottom)
        voxel_size_x : float
            Size of your voxels in the x direction (left to right)
        soma_diameter : float
            The expected in-plane soma diameter (microns)
        ball_xy_size : float
            Elliptical morphological in-plane filter size (microns)
        ball_z_size : float
            Elliptical morphological axial filter size (microns)
        ball_overlap_fraction : float
            Fraction of the morphological filter needed to be filled
            to retain a voxel
        log_sigma_size : float
            Laplacian of Gaussian filter width (as a fraction of soma diameter)
        n_sds_above_mean_thresh : int
            Cell intensity threshold (as a multiple of noise above the mean)
        soma_spread_factor : float
            Cell spread factor (for splitting up cell clusters)
        max_cluster_size : int
            Largest putative cell cluster (in cubic um) where splitting
            should be attempted
        use_pre_trained_weights : bool
            Select to use pre-trained model weights
        trained_model : Optional[Path]
            Trained model file path (home directory (default) -> pretrained
            weights)
        start_plane : int
            First plane to process (to process a subset of the data)
        end_plane : int
            Last plane to process (to process a subset of the data)
        n_free_cpus : int
            How many CPU cores to leave free
        analyse_local : bool
            Only analyse planes around the current position
        debug : bool
            Increase logging
        reset_button :
            Reset parameters to default
        """
        if signal_image is None or background_image is None:
            show_info("Both signal and background images must be specified.")
            return
        data_inputs = DataInputs(
            signal_image.data,
            background_image.data,
            voxel_size_z,
            voxel_size_y,
            voxel_size_x,
        )

        detection_inputs = DetectionInputs(
            soma_diameter,
            ball_xy_size,
            ball_z_size,
            ball_overlap_fraction,
            log_sigma_size,
            n_sds_above_mean_thresh,
            soma_spread_factor,
            max_cluster_size,
        )

        if use_pre_trained_weights:
            trained_model = None
        classification_inputs = ClassificationInputs(
            use_pre_trained_weights, trained_model
        )

        end_plane = len(signal_image.data) if end_plane == 0 else end_plane

        if analyse_local:
            current_plane = viewer.dims.current_step[0]

            # so a reasonable number of cells in the plane are detected
            planes_needed = MIN_PLANES_ANALYSE + int(
                ceil((CUBE_DEPTH * NETWORK_VOXEL_SIZES[0]) / voxel_size_z)
            )

            start_plane, end_plane = get_cube_depth_min_max(
                current_plane, planes_needed
            )
            start_plane = max(0, start_plane)
            end_plane = min(len(signal_image.data), end_plane)

        misc_inputs = MiscInputs(
            start_plane, end_plane, n_free_cpus, analyse_local, debug
        )

        worker = Worker(
            data_inputs,
            detection_inputs,
            classification_inputs,
            misc_inputs,
        )
        worker.returned.connect(
            lambda points: add_layers(points, viewer=viewer)
        )

        # Make sure if the worker emits an error, it is propagated to this
        # thread
        def reraise(e):
            raise Exception from e

        worker.errored.connect(reraise)

        def update_progress_bar(label: str, max: int, value: int):
            progress_bar.label = label
            progress_bar.max = max
            progress_bar.value = value

        worker.update_progress_bar.connect(update_progress_bar)
        worker.start()

    widget.header.value = widget_header
    widget.header.native.setOpenExternalLinks(True)

    @widget.reset_button.changed.connect
    def restore_defaults():
        """
        Restore default widget values.
        """
        defaults = {
            **DataInputs.defaults(),
            **DetectionInputs.defaults(),
            **ClassificationInputs.defaults(),
            **MiscInputs.defaults(),
        }
        for name, value in defaults.items():
            if value is not None:  # ignore fields with no default
                getattr(widget, name).value = value

    # Insert progress bar before the run and reset buttons
    widget.insert(-3, progress_bar)

    scroll = QScrollArea()
    scroll.setWidget(widget._widget._qwidget)
    widget._widget._qwidget = scroll

    return widget



import math
import os
from queue import Queue
from threading import Lock
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from brainglobe_utils.cells.cells import Cell
from tifffile import tifffile
from tqdm import tqdm

from cellfinder.core import logger
from cellfinder.core.detect.filters.setup_filters import (
    get_ball_filter,
    get_cell_detector,
)
from cellfinder.core.detect.filters.volume.structure_detection import (
    get_structure_centre,
)
from cellfinder.core.detect.filters.volume.structure_splitting import (
    StructureSplitException,
    split_cells,
)


class VolumeFilter(object):
    def __init__(
        self,
        *,
        soma_diameter: float,
        soma_size_spread_factor: float = 1.4,
        setup_params: Tuple[np.ndarray, Any, int, int, float, Any],
        n_planes: int,
        n_locks_release: int,
        save_planes: bool = False,
        plane_directory: Optional[str] = None,
        start_plane: int = 0,
        max_cluster_size: int = 5000,
        outlier_keep: bool = False,
        artifact_keep: bool = True,
    ):
        self.soma_diameter = soma_diameter
        self.soma_size_spread_factor = soma_size_spread_factor
        self.n_planes = n_planes
        self.z = start_plane
        self.save_planes = save_planes
        self.plane_directory = plane_directory
        self.max_cluster_size = max_cluster_size
        self.outlier_keep = outlier_keep
        self.n_locks_release = n_locks_release

        self.artifact_keep = artifact_keep

        self.clipping_val = None
        self.threshold_value = None
        self.setup_params = setup_params

        self.previous_plane: Optional[np.ndarray] = None

        self.ball_filter = get_ball_filter(
            plane=self.setup_params[0],
            soma_diameter=self.setup_params[1],
            ball_xy_size=self.setup_params[2],
            ball_z_size=self.setup_params[3],
            ball_overlap_fraction=self.setup_params[4],
        )

        self.cell_detector = get_cell_detector(
            plane_shape=self.setup_params[0].shape,  # type: ignore
            ball_z_size=self.setup_params[3],
            z_offset=self.setup_params[5],
        )

    def process(
        self,
        async_result_queue: Queue,
        locks: List[Lock],
        *,
        callback: Callable[[int], None],
    ) -> List[Cell]:
        progress_bar = tqdm(total=self.n_planes, desc="Processing planes")
        for z in range(self.n_planes):
            # Get result from the queue.
            #
            # It is important to remove the result from the queue here
            # to free up memory once this plane has been processed by
            # the 3D filter here
            logger.debug(f"🏐 Waiting for plane {z}")
            result = async_result_queue.get()
            # .get() blocks until the result is available
            plane, mask = result.get()
            logger.debug(f"🏐 Got plane {z}")

            self.ball_filter.append(plane, mask)

            if self.ball_filter.ready:
                # Let the next 2D filter run
                z_release = z + self.n_locks_release + 1
                if z_release < len(locks):
                    logger.debug(f"🔓 Releasing lock for plane {z_release}")
                    locks[z_release].release()

                self._run_filter()

            callback(self.z)
            self.z += 1
            progress_bar.update()

        progress_bar.close()
        logger.debug("3D filter done")
        return self.get_results()

    def _run_filter(self) -> None:
        logger.debug(f"🏐 Ball filtering plane {self.z}")
        self.ball_filter.walk()

        middle_plane = self.ball_filter.get_middle_plane()
        if self.save_planes:
            self.save_plane(middle_plane)

        logger.debug(f"🏫 Detecting structures for plane {self.z}")
        self.previous_plane = self.cell_detector.process(
            middle_plane, self.previous_plane
        )

        logger.debug(f"🏫 Structures done for plane {self.z}")

    def save_plane(self, plane: np.ndarray) -> None:
        if self.plane_directory is None:
            raise ValueError(
                "plane_directory must be set to save planes to file"
            )
        plane_name = f"plane_{str(self.z).zfill(4)}.tif"
        f_path = os.path.join(self.plane_directory, plane_name)
        tifffile.imsave(f_path, plane.T)

    def get_results(self) -> List[Cell]:
        logger.info("Splitting cell clusters and writing results")

        max_cell_volume = sphere_volume(
            self.soma_size_spread_factor * self.soma_diameter / 2
        )

        cells = []

        logger.debug(
            f"Processing {len(self.cell_detector.coords_maps.items())} cells"
        )
        for cell_id, cell_points in self.cell_detector.coords_maps.items():
            cell_volume = len(cell_points)

            if cell_volume < max_cell_volume:
                cell_centre = get_structure_centre(cell_points)
                cells.append(
                    Cell(
                        (
                            cell_centre[0],
                            cell_centre[1],
                            cell_centre[2],
                        ),
                        Cell.UNKNOWN,
                    )
                )
            else:
                if cell_volume < self.max_cluster_size:
                    try:
                        cell_centres = split_cells(
                            cell_points, outlier_keep=self.outlier_keep
                        )
                    except (ValueError, AssertionError) as err:
                        raise StructureSplitException(
                            f"Cell {cell_id}, error; {err}"
                        )
                    for cell_centre in cell_centres:
                        cells.append(
                            Cell(
                                (
                                    cell_centre[0],
                                    cell_centre[1],
                                    cell_centre[2],
                                ),
                                Cell.UNKNOWN,
                            )
                        )
                else:
                    cell_centre = get_structure_centre(cell_points)
                    cells.append(
                        Cell(
                            (
                                cell_centre[0],
                                cell_centre[1],
                                cell_centre[2],
                            ),
                            Cell.ARTIFACT,
                        )
                    )

        logger.debug("Finished splitting cell clusters.")
        return cells


def sphere_volume(radius: float) -> float:
    return (4 / 3) * math.pi * radius**3



from typing import List

import numpy as np
import pooch
from napari.types import LayerData
from skimage.io import imread

base_url = (
    "https://raw.githubusercontent.com/brainglobe/cellfinder/"
    "master/tests/data/integration/detection/crop_planes"
)


def load_sample() -> List[LayerData]:
    """
    Load some sample data.
    """
    layers = []
    for ch, name in zip([1, 0], ["Background", "Signal"]):
        data = []
        for i in range(30):
            url = f"{base_url}/ch{ch}/ch{ch}{str(i).zfill(4)}.tif"
            file = pooch.retrieve(url=url, known_hash=None)
            data.append(imread(file))

        data = np.stack(data, axis=0)
        layers.append((data, {"name": name}))

    return layers



import glob
import os
from typing import Tuple

import numpy as np
from brainglobe_utils.general.system import get_sorted_file_paths
from dask import array as da
from dask import delayed
from tifffile import TiffFile, imread


def get_tiff_meta(
    path: str,
) -> Tuple[Tuple[int, int], np.dtype]:
    with TiffFile(path) as tfile:
        nz = len(tfile.pages)
        if not nz:
            raise ValueError(f"tiff file {path} has no pages!")
        first_page = tfile.pages[0]

    return tfile.pages[0].shape, first_page.dtype


lazy_imread = delayed(imread)  # lazy reader


def read_with_dask(path):
    """
    Based on https://github.com/tlambert03/napari-ndtiffs
    :param path:
    :return:
    """

    if path.endswith(".txt"):
        with open(path, "r") as f:
            filenames = [line.rstrip() for line in f.readlines()]

    else:
        filenames = glob.glob(os.path.join(path, "*.tif"))

    shape, dtype = get_tiff_meta(filenames[0])
    lazy_arrays = [lazy_imread(fn) for fn in get_sorted_file_paths(filenames)]
    dask_arrays = [
        da.from_delayed(delayed_reader, shape=shape, dtype=dtype)
        for delayed_reader in lazy_arrays
    ]
    stack = da.stack(dask_arrays, axis=0)
    return stack



"""
prep
==================
Functions to prepare files and directories needed for other functions
"""

import os
from pathlib import Path
from typing import Optional

from brainglobe_utils.general.config import get_config_obj
from brainglobe_utils.general.system import get_num_processes

import cellfinder.core.tools.tf as tf_tools
from cellfinder.core import logger
from cellfinder.core.download import models as model_download
from cellfinder.core.download.download import amend_cfg
from cellfinder.core.tools.source_files import source_custom_config_cellfinder

home = Path.home()
DEFAULT_INSTALL_PATH = home / ".cellfinder"


def prep_model_weights(
    model_weights: Optional[os.PathLike],
    install_path: Optional[os.PathLike],
    model_name: model_download.model_type,
    n_free_cpus: int,
) -> Path:
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    prep_tensorflow(n_processes)
    model_weights = prep_models(model_weights, install_path, model_name)

    return model_weights


def prep_tensorflow(max_threads: int) -> None:
    tf_tools.set_tf_threads(max_threads)
    tf_tools.allow_gpu_memory_growth()


def prep_models(
    model_weights_path: Optional[os.PathLike],
    install_path: Optional[os.PathLike],
    model_name: model_download.model_type,
) -> Path:
    install_path = install_path or DEFAULT_INSTALL_PATH
    # if no model or weights, set default weights
    if model_weights_path is None:
        logger.debug("No model supplied, so using the default")

        config_file = source_custom_config_cellfinder()

        if not Path(config_file).exists():
            logger.debug("Custom config does not exist, downloading models")
            model_path = model_download.main(model_name, install_path)
            amend_cfg(new_model_path=model_path)

        model_weights = get_model_weights(config_file)
        if not model_weights.exists():
            logger.debug("Model weights do not exist, downloading")
            model_path = model_download.main(model_name, install_path)
            amend_cfg(new_model_path=model_path)
            model_weights = get_model_weights(config_file)
    else:
        model_weights = Path(model_weights_path)
    return model_weights


def get_model_weights(config_file: os.PathLike) -> Path:
    logger.debug(f"Reading config file: {config_file}")
    config_obj = get_config_obj(config_file)
    model_conf = config_obj["model"]
    model_weights = model_conf["model_path"]
    return Path(model_weights)



"""
main
===============

Trains a network based on a yaml file specifying cubes of cells/non cells.

N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""

import os
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    ArgumentTypeError,
)
from datetime import datetime
from pathlib import Path
from typing import Dict, Literal

from brainglobe_utils.general.numerical import (
    check_positive_float,
    check_positive_int,
)
from brainglobe_utils.general.system import ensure_directory_exists
from brainglobe_utils.IO.cells import find_relevant_tiffs
from brainglobe_utils.IO.yaml import read_yaml_section
from fancylog import fancylog
from sklearn.model_selection import train_test_split

import cellfinder.core as program_for_log
from cellfinder.core import logger
from cellfinder.core.classify.resnet import layer_type
from cellfinder.core.tools.prep import DEFAULT_INSTALL_PATH

tf_suppress_log_messages = [
    "sample_weight modes were coerced from",
    "multiprocessing can interact badly with TensorFlow",
]

depth_type = Literal["18", "34", "50", "101", "152"]

models: Dict[depth_type, layer_type] = {
    "18": "18-layer",
    "34": "34-layer",
    "50": "50-layer",
    "101": "101-layer",
    "152": "152-layer",
}


def valid_model_depth(depth):
    """
    Ensures a correct existing_model is chosen
    :param value: Input value
    :param models: Dict of allowed models
    :return: Input value, if it corresponds to a valid existing_model
    """

    if depth in models.keys():
        return depth
    else:
        raise ArgumentTypeError(
            f"Model depth: {depth} is not valid. Please "
            f"choose one of: {list(models.keys())}"
        )


def misc_parse(parser):
    misc_parser = parser.add_argument_group("Misc options")
    misc_parser.add_argument(
        "--n-free-cpus",
        dest="n_free_cpus",
        type=check_positive_int,
        default=2,
        help="The number of CPU cores on the machine to leave "
        "unused by the program to spare resources.",
    )
    misc_parser.add_argument(
        "--max-ram",
        dest="max_ram",
        type=check_positive_float,
        default=None,
        help="Maximum amount of RAM to use (in GB) - not currently fully "
        "implemented for all parts of cellfinder",
    )
    misc_parser.add_argument(
        "--save-csv",
        dest="save_csv",
        action="store_true",
        help="Save .csv files of cell locations (in addition to xml)."
        "Useful for importing into other software.",
    )
    misc_parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Debug mode. Will increase verbosity of logging and save all "
        "intermediate files for diagnosis of software issues.",
    )
    misc_parser.add_argument(
        "--sort-input-file",
        dest="sort_input_file",
        action="store_true",
        help="If set to true, the input text file will be sorted using "
        "natural sorting. This means that the file paths will be "
        "sorted as would be expected by a human and "
        "not purely alphabetically",
    )
    return parser


def training_parse():
    from cellfinder.core.download.cli import (
        download_directory_parser,
        model_parser,
    )

    training_parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter
    )
    training_parser.add_argument(
        "-y",
        "--yaml",
        dest="yaml_file",
        nargs="+",
        required=True,
        type=str,
        help="The path to the yaml run file.",
    )
    training_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        required=True,
        type=str,
        help="Output directory for the final model.",
    )
    training_parser.add_argument(
        "--continue-training",
        dest="continue_training",
        action="store_true",
        help="Continue training from an existing trained model. If no model "
        "or model weights are specified, this will continue from the "
        "included model.",
    )
    training_parser.add_argument(
        "--trained-model",
        dest="trained_model",
        type=str,
        help="Path to the trained model",
    )
    training_parser.add_argument(
        "--model-weights",
        dest="model_weights",
        type=str,
        help="Path to existing model weights",
    )
    training_parser.add_argument(
        "--network-depth",
        dest="network_depth",
        type=valid_model_depth,
        default="50",
        help="Resnet depth (based on He et al. (2015)",
    )
    training_parser.add_argument(
        "--batch-size",
        dest="batch_size",
        type=check_positive_int,
        default=16,
        help="Training batch size",
    )
    training_parser.add_argument(
        "--epochs",
        dest="epochs",
        type=check_positive_int,
        default=100,
        help="Number of training epochs",
    )
    training_parser.add_argument(
        "--test-fraction",
        dest="test_fraction",
        type=float,
        default=0.1,
        help="Fraction of training data to use for validation",
    )
    training_parser.add_argument(
        "--learning-rate",
        dest="learning_rate",
        type=check_positive_float,
        default=0.0001,
        help="Learning rate for training the model",
    )
    training_parser.add_argument(
        "--no-augment",
        dest="no_augment",
        action="store_true",
        help="Don't apply data augmentation",
    )
    training_parser.add_argument(
        "--save-weights",
        dest="save_weights",
        action="store_true",
        help="Only store the model weights, and not the full model. Useful to "
        "save storage space.",
    )
    training_parser.add_argument(
        "--no-save-checkpoints",
        dest="no_save_checkpoints",
        action="store_true",
        help="Store the model at intermediate points during training",
    )
    training_parser.add_argument(
        "--tensorboard",
        action="store_true",
        help="Log to output_directory/tensorboard",
    )
    training_parser.add_argument(
        "--save-progress",
        dest="save_progress",
        action="store_true",
        help="Save training progress to a .csv file",
    )

    training_parser = misc_parse(training_parser)
    training_parser = model_parser(training_parser)
    training_parser = download_directory_parser(training_parser)
    args = training_parser.parse_args()

    return args


def parse_yaml(yaml_files, section="data"):
    data = []
    for yaml_file in yaml_files:
        data.extend(read_yaml_section(yaml_file, section))
    return data


def get_tiff_files(yaml_contents):
    from cellfinder.core.tools.tiff import TiffDir, TiffList

    tiff_lists = []
    for d in yaml_contents:
        if d["bg_channel"] < 0:
            channels = [d["signal_channel"]]
        else:
            channels = [d["signal_channel"], d["bg_channel"]]
        if "cell_def" in d and d["cell_def"]:
            ch1_tiffs = [
                os.path.join(d["cube_dir"], f)
                for f in os.listdir(d["cube_dir"])
                if f.endswith("Ch" + str(channels[0]) + ".tif")
            ]
            tiff_lists.append(
                TiffList(
                    find_relevant_tiffs(ch1_tiffs, d["cell_def"]),
                    channels,
                    d["type"],
                )
            )
        else:
            tiff_lists.append(TiffDir(d["cube_dir"], channels, d["type"]))

    tiff_files = [tiff_dir.make_tifffile_list() for tiff_dir in tiff_lists]
    return tiff_files


def cli():
    args = training_parse()
    ensure_directory_exists(args.output_dir)

    fancylog.start_logging(
        args.output_dir,
        program_for_log,
        variables=[args],
        log_header="CELLFINDER TRAINING LOG",
    )

    output_dir = Path(args.output_dir)
    run(
        output_dir,
        args.yaml_file,
        n_free_cpus=args.n_free_cpus,
        trained_model=args.trained_model,
        model_weights=args.model_weights,
        install_path=args.install_path,
        model=args.model,
        network_depth=args.network_depth,
        learning_rate=args.learning_rate,
        continue_training=args.continue_training,
        test_fraction=args.test_fraction,
        batch_size=args.batch_size,
        no_augment=args.no_augment,
        tensorboard=args.tensorboard,
        save_weights=args.save_weights,
        no_save_checkpoints=args.no_save_checkpoints,
        save_progress=args.save_progress,
        epochs=args.epochs,
    )


def run(
    output_dir,
    yaml_file,
    n_free_cpus=2,
    trained_model=None,
    model_weights=None,
    install_path=DEFAULT_INSTALL_PATH,
    model="resnet50_tv",
    network_depth="50",
    learning_rate=0.0001,
    continue_training=False,
    test_fraction=0.1,
    batch_size=16,
    no_augment=False,
    tensorboard=False,
    save_weights=False,
    no_save_checkpoints=False,
    save_progress=False,
    epochs=100,
):
    from cellfinder.core.main import suppress_tf_logging

    suppress_tf_logging(tf_suppress_log_messages)

    from tensorflow.keras.callbacks import (
        CSVLogger,
        ModelCheckpoint,
        TensorBoard,
    )

    from cellfinder.core.classify.cube_generator import CubeGeneratorFromDisk
    from cellfinder.core.classify.tools import get_model, make_lists
    from cellfinder.core.tools.prep import prep_model_weights

    start_time = datetime.now()

    ensure_directory_exists(output_dir)
    model_weights = prep_model_weights(
        model_weights=model_weights,
        install_path=install_path,
        model_name=model,
        n_free_cpus=n_free_cpus,
    )

    yaml_contents = parse_yaml(yaml_file)

    tiff_files = get_tiff_files(yaml_contents)
    logger.info(
        f"Found {sum(len(imlist) for imlist in tiff_files)} images "
        f"from {len(yaml_contents)} datasets "
        f"in {len(yaml_file)} yaml files"
    )

    model = get_model(
        existing_model=trained_model,
        model_weights=model_weights,
        network_depth=models[network_depth],
        learning_rate=learning_rate,
        continue_training=continue_training,
    )

    signal_train, background_train, labels_train = make_lists(tiff_files)

    if test_fraction > 0:
        logger.info("Splitting data into training and validation datasets")
        (
            signal_train,
            signal_test,
            background_train,
            background_test,
            labels_train,
            labels_test,
        ) = train_test_split(
            signal_train,
            background_train,
            labels_train,
            test_size=test_fraction,
        )

        logger.info(
            f"Using {len(signal_train)} images for training and "
            f"{len(signal_test)} images for validation"
        )
        validation_generator = CubeGeneratorFromDisk(
            signal_test,
            background_test,
            labels=labels_test,
            batch_size=batch_size,
            train=True,
        )

        # for saving checkpoints
        base_checkpoint_file_name = "-epoch.{epoch:02d}-loss-{val_loss:.3f}.h5"

    else:
        logger.info("No validation data selected.")
        validation_generator = None
        base_checkpoint_file_name = "-epoch.{epoch:02d}.h5"

    training_generator = CubeGeneratorFromDisk(
        signal_train,
        background_train,
        labels=labels_train,
        batch_size=batch_size,
        shuffle=True,
        train=True,
        augment=not no_augment,
    )
    callbacks = []

    if tensorboard:
        logdir = output_dir / "tensorboard"
        ensure_directory_exists(logdir)
        tensorboard = TensorBoard(
            log_dir=logdir,
            histogram_freq=0,
            write_graph=True,
            update_freq="epoch",
        )
        callbacks.append(tensorboard)

    if not no_save_checkpoints:
        if save_weights:
            filepath = str(output_dir / ("weight" + base_checkpoint_file_name))
        else:
            filepath = str(output_dir / ("model" + base_checkpoint_file_name))

        checkpoints = ModelCheckpoint(
            filepath,
            save_weights_only=save_weights,
        )
        callbacks.append(checkpoints)

    if save_progress:
        filepath = str(output_dir / "training.csv")
        csv_logger = CSVLogger(filepath)
        callbacks.append(csv_logger)

    logger.info("Beginning training.")
    model.fit(
        training_generator,
        validation_data=validation_generator,
        use_multiprocessing=False,
        epochs=epochs,
        callbacks=callbacks,
    )

    if save_weights:
        logger.info("Saving model weights")
        model.save_weights(str(output_dir / "model_weights.h5"))
    else:
        logger.info("Saving model")
        model.save(output_dir / "model.h5")

    logger.info(
        "Finished training, " "Total time taken: %s",
        datetime.now() - start_time,
    )


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from threading import Lock
from typing import Optional, Tuple

import dask.array as da
import numpy as np

from cellfinder.core import types
from cellfinder.core.detect.filters.plane.classical_filter import enhance_peaks
from cellfinder.core.detect.filters.plane.tile_walker import TileWalker


@dataclass
class TileProcessor:
    """
    Attributes
    ----------
    clipping_value :
        Upper value that the input plane is clipped to.
    threshold_value :
        Value used to mark bright features in the input planes after they have
        been run through the 2D filter.
    """

    clipping_value: int
    threshold_value: int
    soma_diameter: int
    log_sigma_size: float
    n_sds_above_mean_thresh: float

    def get_tile_mask(
        self, plane: types.array, lock: Optional[Lock] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This thresholds the input plane, and returns a mask indicating which
        tiles are inside the brain.

        The input plane is:

        1. Clipped to [0, self.clipping_value]
        2. Run through a peak enhancement filter (see `classical_filter.py`)
        3. Thresholded. Any values that are larger than
           (mean + stddev * self.n_sds_above_mean_thresh) are set to
           self.threshold_value in-place.

        Parameters
        ----------
        plane :
            Input plane.
        lock :
            If given, block reading the plane into memory until the lock
            can be acquired.

        Returns
        -------
        plane :
            Thresholded plane.
        inside_brain_tiles :
            Boolean mask indicating which tiles are inside (1) or
            outside (0) the brain.
        """
        laplace_gaussian_sigma = self.log_sigma_size * self.soma_diameter
        plane = plane.T
        np.clip(plane, 0, self.clipping_value, out=plane)
        if lock is not None:
            lock.acquire(blocking=True)
        # Read plane from a dask array into memory as a numpy array
        if isinstance(plane, da.Array):
            plane = np.array(plane)

        # Get tiles that are within the brain
        walker = TileWalker(plane, self.soma_diameter)
        walker.mark_bright_tiles()
        inside_brain_tiles = walker.bright_tiles_mask

        # Threshold the image
        thresholded_img = enhance_peaks(
            plane.copy(),
            self.clipping_value,
            gaussian_sigma=laplace_gaussian_sigma,
        )
        avg = np.mean(thresholded_img)
        sd = np.std(thresholded_img)
        threshold = avg + self.n_sds_above_mean_thresh * sd
        plane[thresholded_img > threshold] = self.threshold_value

        return plane, inside_brain_tiles



"""
Detection is run in three steps:

1. 2D filtering
2. 3D filtering
3. Structure detection

In steps 1. and 2. filters are applied, and any bright points detected
post-filter are marked. To avoid using a separate mask array to mark the
bright points, the input data is clipped to [0, (max_val - 2)]
(max_val is the maximum value that the image data type can store), and:
- (max_val - 1) is used to mark bright points during 2D filtering
- (max_val) is used to mark bright points during 3D filtering
"""

import multiprocessing
from datetime import datetime
from queue import Queue
from threading import Lock
from typing import Callable, List, Optional, Sequence, Tuple, TypeVar

import numpy as np
from brainglobe_utils.cells.cells import Cell
from brainglobe_utils.general.system import get_num_processes

from cellfinder.core import logger, types
from cellfinder.core.detect.filters.plane import TileProcessor
from cellfinder.core.detect.filters.setup_filters import setup_tile_filtering
from cellfinder.core.detect.filters.volume.volume_filter import VolumeFilter


def calculate_parameters_in_pixels(
    voxel_sizes: Tuple[float, float, float],
    soma_diameter_um: float,
    max_cluster_size_um3: float,
    ball_xy_size_um: float,
    ball_z_size_um: float,
) -> Tuple[int, int, int, int]:
    """
    Convert the command-line arguments from real (um) units to pixels
    """

    mean_in_plane_pixel_size = 0.5 * (
        float(voxel_sizes[2]) + float(voxel_sizes[1])
    )
    voxel_volume = (
        float(voxel_sizes[2]) * float(voxel_sizes[1]) * float(voxel_sizes[0])
    )
    soma_diameter = int(round(soma_diameter_um / mean_in_plane_pixel_size))
    max_cluster_size = int(round(max_cluster_size_um3 / voxel_volume))
    ball_xy_size = int(round(ball_xy_size_um / mean_in_plane_pixel_size))
    ball_z_size = int(round(ball_z_size_um / float(voxel_sizes[0])))

    if ball_z_size == 0:
        raise ValueError(
            "Ball z size has been calculated to be 0 voxels."
            " This may be due to large axial spacing of your data or the "
            "ball_z_size_um parameter being too small. "
            "Please check input parameters are correct. "
            "Note that cellfinder requires high resolution data in all "
            "dimensions, so that cells can be detected in multiple "
            "image planes."
        )
    return soma_diameter, max_cluster_size, ball_xy_size, ball_z_size


def main(
    signal_array: types.array,
    start_plane: int,
    end_plane: int,
    voxel_sizes: Tuple[float, float, float],
    soma_diameter: float,
    max_cluster_size: float,
    ball_xy_size: float,
    ball_z_size: float,
    ball_overlap_fraction: float,
    soma_spread_factor: float,
    n_free_cpus: int,
    log_sigma_size: float,
    n_sds_above_mean_thresh: float,
    outlier_keep: bool = False,
    artifact_keep: bool = False,
    save_planes: bool = False,
    plane_directory: Optional[str] = None,
    *,
    callback: Optional[Callable[[int], None]] = None,
) -> List[Cell]:
    """
    Perform cell candidate detection on a 3D signal array.

    Parameters
    ----------
    signal_array : numpy.ndarray
        3D array representing the signal data.

    start_plane : int
        Index of the starting plane for detection.

    end_plane : int
        Index of the ending plane for detection.

    voxel_sizes : Tuple[float, float, float]
        Tuple of voxel sizes in each dimension (x, y, z).

    soma_diameter : float
        Diameter of the soma in physical units.

    max_cluster_size : float
        Maximum size of a cluster in physical units.

    ball_xy_size : float
        Size of the XY ball used for filtering in physical units.

    ball_z_size : float
        Size of the Z ball used for filtering in physical units.

    ball_overlap_fraction : float
        Fraction of overlap allowed between balls.

    soma_spread_factor : float
        Spread factor for soma size.

    n_free_cpus : int
        Number of free CPU cores available for parallel processing.

    log_sigma_size : float
        Size of the sigma for the log filter.

    n_sds_above_mean_thresh : float
        Number of standard deviations above the mean threshold.

    outlier_keep : bool, optional
        Whether to keep outliers during detection. Defaults to False.

    artifact_keep : bool, optional
        Whether to keep artifacts during detection. Defaults to False.

    save_planes : bool, optional
        Whether to save the planes during detection. Defaults to False.

    plane_directory : str, optional
        Directory path to save the planes. Defaults to None.

    callback : Callable[int], optional
        A callback function that is called every time a plane has finished
        being processed. Called with the plane number that has finished.

    Returns
    -------
    List[Cell]
        List of detected cells.
    """
    if not np.issubdtype(signal_array.dtype, np.integer):
        raise ValueError(
            "signal_array must be integer datatype, but has datatype "
            f"{signal_array.dtype}"
        )
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    n_ball_procs = max(n_processes - 1, 1)
    start_time = datetime.now()

    (
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
    ) = calculate_parameters_in_pixels(
        voxel_sizes,
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
    )

    if end_plane == -1:
        end_plane = len(signal_array)
    signal_array = signal_array[start_plane:end_plane]
    signal_array = signal_array.astype(np.uint32)

    callback = callback or (lambda *args, **kwargs: None)

    if signal_array.ndim != 3:
        raise ValueError("Input data must be 3D")

    setup_params = (
        signal_array[0, :, :],
        soma_diameter,
        ball_xy_size,
        ball_z_size,
        ball_overlap_fraction,
        start_plane,
    )

    # Create 3D analysis filter
    mp_3d_filter = VolumeFilter(
        soma_diameter=soma_diameter,
        setup_params=setup_params,
        soma_size_spread_factor=soma_spread_factor,
        n_planes=len(signal_array),
        n_locks_release=n_ball_procs,
        save_planes=save_planes,
        plane_directory=plane_directory,
        start_plane=start_plane,
        max_cluster_size=max_cluster_size,
        outlier_keep=outlier_keep,
        artifact_keep=artifact_keep,
    )

    clipping_val, threshold_value = setup_tile_filtering(signal_array[0, :, :])
    # Create 2D analysis filter
    mp_tile_processor = TileProcessor(
        clipping_val,
        threshold_value,
        soma_diameter,
        log_sigma_size,
        n_sds_above_mean_thresh,
    )

    # Force spawn context
    mp_ctx = multiprocessing.get_context("spawn")
    with mp_ctx.Pool(n_ball_procs) as worker_pool:
        async_results, locks = _map_with_locks(
            mp_tile_processor.get_tile_mask,
            signal_array,  # type: ignore
            worker_pool,
        )

        # Release the first set of locks for the 2D filtering
        for i in range(min(n_ball_procs + ball_z_size, len(locks))):
            logger.debug(f"🔓 Releasing lock for plane {i}")
            locks[i].release()

        # Start 3D filter
        #
        # This runs in the main thread, and blocks until the all the 2D and
        # then 3D filtering has finished. As batches of planes are filtered
        # by the 3D filter, it releases the locks of subsequent 2D filter
        # processes.
        cells = mp_3d_filter.process(async_results, locks, callback=callback)

    time_elapsed = datetime.now() - start_time
    logger.debug(
        f"All Planes done. Found {len(cells)} cells in {format(time_elapsed)}"
    )
    print("Detection complete - all planes done in : {}".format(time_elapsed))
    return cells


Tin = TypeVar("Tin")
Tout = TypeVar("Tout")


def _run_func_with_lock(
    func: Callable[[Tin], Tout], arg: Tin, lock: Lock
) -> Tout:
    """
    Run a function after acquiring a lock.
    """
    lock.acquire(blocking=True)
    return func(arg)


def _map_with_locks(
    func: Callable[[Tin], Tout],
    iterable: Sequence[Tin],
    worker_pool: multiprocessing.pool.Pool,
) -> Tuple[Queue, List[Lock]]:
    """
    Map a function to arguments, blocking execution.

    Maps *func* to args in *iterable*, but blocks all execution and
    return a queue of asyncronous results and locks for each of the
    results. Execution can be enabled by releasing the returned
    locks in order.
    """
    # Setup a manager to handle the locks
    m = multiprocessing.Manager()
    # Setup one lock per argument to be mapped
    locks = [m.Lock() for _ in range(len(iterable))]
    [lock.acquire(blocking=False) for lock in locks]

    async_results: Queue = Queue()

    for arg, lock in zip(iterable, locks):
        async_result = worker_pool.apply_async(
            _run_func_with_lock, args=(func, arg, lock)
        )
        async_results.put(async_result)

    return async_results, locks



import math
from typing import Tuple

import numpy as np

from cellfinder.core.detect.filters.volume.ball_filter import BallFilter
from cellfinder.core.detect.filters.volume.structure_detection import (
    CellDetector,
)
from cellfinder.core.tools.tools import get_max_possible_value


def get_ball_filter(
    *,
    plane: np.ndarray,
    soma_diameter: int,
    ball_xy_size: int,
    ball_z_size: int,
    ball_overlap_fraction: float = 0.6,
) -> BallFilter:
    # thrsh_val is used to clip the data in plane to make sure
    # a number is available to mark cells. soma_centre_val is the
    # number used to mark cells.
    max_value = get_max_possible_value(plane)
    thrsh_val = max_value - 1
    soma_centre_val = max_value

    tile_width = soma_diameter * 2
    plane_height, plane_width = plane.shape

    ball_filter = BallFilter(
        plane_width,
        plane_height,
        ball_xy_size,
        ball_z_size,
        overlap_fraction=ball_overlap_fraction,
        tile_step_width=tile_width,
        tile_step_height=tile_width,
        threshold_value=thrsh_val,
        soma_centre_value=soma_centre_val,
    )
    return ball_filter


def get_cell_detector(
    *, plane_shape: Tuple[int, int], ball_z_size: int, z_offset: int = 0
) -> CellDetector:
    plane_height, plane_width = plane_shape
    start_z = z_offset + int(math.floor(ball_z_size / 2))
    return CellDetector(plane_width, plane_height, start_z=start_z)


def setup_tile_filtering(plane: np.ndarray) -> Tuple[int, int]:
    """
    Setup values that are used to threshold the plane during 2D filtering.

    Returns
    -------
    clipping_value :
        Upper value used to clip planes before 2D filtering. This is chosen
        to leave two numbers left that can later be used to mark bright points
        during the 2D and 3D filtering stages.
    threshold_value :
        Value used to mark bright pixels after 2D filtering.
    """
    max_value = get_max_possible_value(plane)
    clipping_value = max_value - 2
    thrsh_val = max_value - 1

    return clipping_value, thrsh_val



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("cellfinder")
except PackageNotFoundError as e:
    raise PackageNotFoundError("cellfinder package not installed") from e

# If tensorflow is not present, tools cannot be used.
# Throw an error in this case to prevent invocation of functions.
try:
    TF_VERSION = version("tensorflow")
except PackageNotFoundError as e:
    try:
        TF_VERSION = version("tensorflow-macos")
    except PackageNotFoundError as e:
        raise PackageNotFoundError(
            f"cellfinder tools cannot be invoked without tensorflow. "
            f"Please install tensorflow into your environment to use cellfinder tools. "
            f"For more information, please see "
            f"https://github.com/brainglobe/brainglobe-meta#readme."
        ) from e

__author__ = "Adam Tyson, Christian Niedworok, Charly Rousseau"
__license__ = "BSD-3-Clause"



import math

import numpy as np


def get_2d_bins(
    arr: np.ndarray, bin_width: int, bin_height: int
) -> np.ndarray:
    """
    Bin a 2D array.

    Parameters
    ----------
    arr :
        Input array.
    bin_width, bin_height :
        Width/height of the bins.

    Returns
    -------
    binned_arr
        A 4D array. The dimensions correspond to:
            1. Bin number along first dimension of input array.
            2. Element index along first dimension within bin.
            3. Bin number along second dimension of input array.
            4. Element index along second dimension within bin.

    Notes
    -----
    The original array is zero padded so that it is tiled exactly
    by tiles of shape (bin_width, bin_height).
    """
    bin_width = int(bin_width)
    bin_height = int(bin_height)

    # Get number of required bins to tile arr
    n_bins_width = int(math.ceil(arr.shape[0] / bin_width))
    n_bins_height = int(math.ceil(arr.shape[1] / bin_height))
    # Get width/height to have integer number of bins that fully
    # covers arr
    width = n_bins_width * bin_width
    height = n_bins_height * bin_height

    # Zero padding array to integer multiples of bin_width/bin_height
    padded_array = np.zeros((width, height), dtype=arr.dtype)
    padded_array[0 : arr.shape[0], 0 : arr.shape[1]] = arr

    # Reshape into 4D array.
    binned_array = padded_array.reshape(
        n_bins_width,
        bin_width,
        n_bins_height,
        bin_height,
    )
    return binned_array


def binned_mean_2d(
    arr: np.ndarray, bin_width: int, bin_height: int
) -> np.ndarray:
    """
    Get mean value of *arr* in bins of shape (*bin_width*, *bin_height*)

    Notes
    -----
    The original array is zero padded so that it is tiled exactly
    by tiles of shape (*bin_width*, *bin_height*). If the tile shape
    does not exactly divide the array shape, values on the outer edge
    of the returned array (arr[-1, :] and arr[:, -1]) will have mean values
    calculated including this zero padding.
    """
    binned_array = get_2d_bins(arr, bin_width, bin_height)
    return binned_array.mean(axis=(1, 3))


def bin_mean_3d(
    arr: np.ndarray, bin_width: int, bin_height: int, bin_depth: int
) -> np.ndarray:
    """
    Bin a 3D array, and return an array of the mean value in each bin.

    Notes
    -----
    If the tile shape does not exactly divide the array shape an error is
    raised.
    """
    if (arr.shape[0] % bin_width) != 0:
        raise ValueError(
            "Bin width must divide arr exactly, got {} and {}".format(
                arr.shape[0], bin_width
            )
        )
    if (arr.shape[1] % bin_height) != 0:
        raise ValueError(
            "Bin height must divide arr exactly, got {} and {}".format(
                arr.shape[1], bin_height
            )
        )
    if (arr.shape[2] % bin_depth) != 0:
        raise ValueError(
            "Bin depth must divide arr exactly, got {} and {}".format(
                arr.shape[2], bin_depth
            )
        )

    binned_arr = []
    for i in range(0, arr.shape[2] - 1, bin_depth):
        sub_stack = [
            binned_mean_2d(arr[:, :, j], bin_width, bin_height)
            for j in range(i, i + bin_depth)
        ]
        binned_arr.append(np.dstack(sub_stack).mean(axis=2))
    return np.dstack(binned_arr)



"""
N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""

import os
from typing import Callable, List, Optional, Tuple

import numpy as np
from brainglobe_utils.general.logging import suppress_specific_logs

from cellfinder.core import logger
from cellfinder.core.download.models import model_type
from cellfinder.core.train.train_yml import depth_type

tf_suppress_log_messages = [
    "multiprocessing can interact badly with TensorFlow"
]


def main(
    signal_array: np.ndarray,
    background_array: np.ndarray,
    voxel_sizes: Tuple[int, int, int],
    start_plane: int = 0,
    end_plane: int = -1,
    trained_model: Optional[os.PathLike] = None,
    model_weights: Optional[os.PathLike] = None,
    model: model_type = "resnet50_tv",
    batch_size: int = 32,
    n_free_cpus: int = 2,
    network_voxel_sizes: Tuple[int, int, int] = (5, 1, 1),
    soma_diameter: int = 16,
    ball_xy_size: int = 6,
    ball_z_size: int = 15,
    ball_overlap_fraction: float = 0.6,
    log_sigma_size: float = 0.2,
    n_sds_above_mean_thresh: int = 10,
    soma_spread_factor: float = 1.4,
    max_cluster_size: int = 100000,
    cube_width: int = 50,
    cube_height: int = 50,
    cube_depth: int = 20,
    network_depth: depth_type = "50",
    *,
    detect_callback: Optional[Callable[[int], None]] = None,
    classify_callback: Optional[Callable[[int], None]] = None,
    detect_finished_callback: Optional[Callable[[list], None]] = None,
) -> List:
    """
    Parameters
    ----------
    detect_callback : Callable[int], optional
        Called every time a plane has finished being processed during the
        detection stage. Called with the plane number that has finished.
    classify_callback : Callable[int], optional
        Called every time tensorflow has finished classifying a point.
        Called with the batch number that has just finished.
    detect_finished_callback : Callable[list], optional
        Called after detection is finished with the list of detected points.
    """
    suppress_tf_logging(tf_suppress_log_messages)

    from cellfinder.core.classify import classify
    from cellfinder.core.detect import detect
    from cellfinder.core.tools import prep

    logger.info("Detecting cell candidates")

    points = detect.main(
        signal_array,
        start_plane,
        end_plane,
        voxel_sizes,
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
        ball_overlap_fraction,
        soma_spread_factor,
        n_free_cpus,
        log_sigma_size,
        n_sds_above_mean_thresh,
        callback=detect_callback,
    )

    if detect_finished_callback is not None:
        detect_finished_callback(points)

    install_path = None
    model_weights = prep.prep_model_weights(
        model_weights, install_path, model, n_free_cpus
    )
    if len(points) > 0:
        logger.info("Running classification")
        points = classify.main(
            points,
            signal_array,
            background_array,
            n_free_cpus,
            voxel_sizes,
            network_voxel_sizes,
            batch_size,
            cube_height,
            cube_width,
            cube_depth,
            trained_model,
            model_weights,
            network_depth,
            callback=classify_callback,
        )
    else:
        logger.info("No candidates, skipping classification")
    return points


def suppress_tf_logging(tf_suppress_log_messages: List[str]) -> None:
    """
    Prevents many lines of logs such as:
    "2019-10-24 16:54:41.363978: I tensorflow/stream_executor/platform/default
    /dso_loader.cc:44] Successfully opened dynamic library libcuda.so.1"
    """
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    for message in tf_suppress_log_messages:
        suppress_specific_logs("tensorflow", message)



from pathlib import Path
from typing import Optional

from magicgui import magicgui
from magicgui.widgets import FunctionGui, PushButton
from napari.qt.threading import thread_worker
from napari.utils.notifications import show_info
from qtpy.QtWidgets import QScrollArea

from cellfinder.core.train.train_yml import run as train_yml
from cellfinder.napari.utils import (
    header_label_widget,
    html_label_widget,
    widget_header,
)

from .train_containers import (
    MiscTrainingInputs,
    OptionalNetworkInputs,
    OptionalTrainingInputs,
    TrainingDataInputs,
)


@thread_worker
def run_training(
    training_data_inputs: TrainingDataInputs,
    optional_network_inputs: OptionalNetworkInputs,
    optional_training_inputs: OptionalTrainingInputs,
    misc_training_inputs: MiscTrainingInputs,
):
    print("Running training")
    train_yml(
        **training_data_inputs.as_core_arguments(),
        **optional_network_inputs.as_core_arguments(),
        **optional_training_inputs.as_core_arguments(),
        **misc_training_inputs.as_core_arguments(),
    )
    print("Finished!")


def training_widget() -> FunctionGui:
    @magicgui(
        header=header_label_widget,
        training_label=html_label_widget("Network training", tag="h3"),
        **TrainingDataInputs.widget_representation(),
        **OptionalNetworkInputs.widget_representation(),
        **OptionalTrainingInputs.widget_representation(),
        **MiscTrainingInputs.widget_representation(),
        call_button=True,
        reset_button=dict(widget_type="PushButton", text="Reset defaults"),
        scrollable=True,
    )
    def widget(
        header: dict,
        training_label: dict,
        data_options: dict,
        yaml_files: Path,
        output_directory: Path,
        network_options: dict,
        trained_model: Optional[Path],
        model_weights: Optional[Path],
        model_depth: str,
        pretrained_model: str,
        training_options: dict,
        continue_training: bool,
        augment: bool,
        tensorboard: bool,
        save_weights: bool,
        save_checkpoints: bool,
        save_progress: bool,
        epochs: int,
        learning_rate: float,
        batch_size: int,
        test_fraction: float,
        misc_options: dict,
        number_of_free_cpus: int,
        reset_button: PushButton,
    ):
        """
        Parameters
        ----------
        yaml_files : Path
            YAML files containing paths to training data
        output_directory : Path
            Directory to save the output trained model
        trained_model : Optional[Path]
            Existing pre-trained model
        model_weights : Optional[Path]
            Existing pre-trained model weights
            Should be set along with "Model depth"
        model_depth : str
            ResNet model depth (as per He et al. (2015))
        pretrained_model : str
            Which pre-trained model to use
            (Supplied with cellfinder)
        continue_training : bool
            Continue training from an existing trained model
            If no trained model or model weights are specified,
            this will continue from the pretrained model
        augment : bool
            Augment the training data to improve generalisation
        tensorboard : bool
            Log to output_directory/tensorboard
        save_weights : bool
            Only store the model weights, and not the full model
            Useful to save storage space
        save_checkpoints : bool
            Store the model at intermediate points during training
        save_progress : bool
            Save training progress to a .csv file
        epochs : int
            Number of training epochs
            (How many times to use each training data point)
        learning_rate : float
            Learning rate for training the model
        batch_size : int
            Training batch size
        test_fraction : float
            Fraction of training data to use for validation
        number_of_free_cpus : int
            How many CPU cores to leave free
        reset_button : PushButton
            Reset parameters to default
        """
        trained_model = None if trained_model == Path.home() else trained_model
        model_weights = None if model_weights == Path.home() else model_weights

        training_data_inputs = TrainingDataInputs(yaml_files, output_directory)

        optional_network_inputs = OptionalNetworkInputs(
            trained_model,
            model_weights,
            model_depth,
            pretrained_model,
        )

        optional_training_inputs = OptionalTrainingInputs(
            continue_training,
            augment,
            tensorboard,
            save_weights,
            save_checkpoints,
            save_progress,
            epochs,
            learning_rate,
            batch_size,
            test_fraction,
        )

        misc_training_inputs = MiscTrainingInputs(number_of_free_cpus)

        if yaml_files[0] == Path.home():  # type: ignore
            show_info("Please select a YAML file for training")
        else:
            worker = run_training(
                training_data_inputs,
                optional_network_inputs,
                optional_training_inputs,
                misc_training_inputs,
            )
            worker.start()

    widget.header.value = widget_header
    widget.header.native.setOpenExternalLinks(True)

    @widget.reset_button.changed.connect
    def restore_defaults():
        defaults = {
            **TrainingDataInputs.defaults(),
            **OptionalNetworkInputs.defaults(),
            **OptionalTrainingInputs.defaults(),
            **MiscTrainingInputs.defaults(),
        }
        for name, value in defaults.items():
            # ignore fields with no default
            if value is not None:
                getattr(widget, name).value = value

    scroll = QScrollArea()
    scroll.setWidget(widget._widget._qwidget)
    widget._widget._qwidget = scroll

    return widget



import os
from pathlib import Path
from typing import Literal

from cellfinder.core import logger
from cellfinder.core.download.download import download

model_weight_urls = {
    "resnet50_tv": "https://gin.g-node.org/cellfinder/models/raw/"
    "master/resnet50_tv.h5",
    "resnet50_all": "https://gin.g-node.org/cellfinder/models/raw/"
    "master/resnet50_weights.h5",
}

download_requirements_gb = {
    "resnet50_tv": 0.18,
    "resnet50_all": 0.18,
}

model_type = Literal["resnet50_tv", "resnet50_all"]


def main(model_name: model_type, download_path: os.PathLike) -> Path:
    """
    For a given model name and download path, download the model file
    and return the path to the downloaded file.
    """
    download_path = Path(download_path)

    model_weight_dir = download_path / "model_weights"
    model_path = model_weight_dir / f"{model_name}.h5"
    if not model_path.exists():
        model_weight_dir.mkdir(parents=True)

        logger.info(
            f"Downloading '{model_name}' model. This may take a little while."
        )

        download(
            model_path,
            model_weight_urls[model_name],
            model_name,
            download_requires=download_requirements_gb[model_name],
        )

    else:
        logger.info(f"Model already exists at {model_path}. Skipping download")

    return model_path



from typing import List, Tuple

import numpy as np

from cellfinder.core import logger
from cellfinder.core.detect.filters.volume.ball_filter import BallFilter
from cellfinder.core.detect.filters.volume.structure_detection import (
    CellDetector,
    get_structure_centre,
)


class StructureSplitException(Exception):
    pass


def get_shape(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> List[int]:
    # +1 because difference. TEST:
    shape = [int((dim.max() - dim.min()) + 1) for dim in (xs, ys, zs)]
    return shape


def coords_to_volume(
    xs: np.ndarray, ys: np.ndarray, zs: np.ndarray, ball_radius: int = 1
) -> np.ndarray:
    ball_diameter = ball_radius * 2
    # Expanded to ensure the ball fits even at the border
    expanded_shape = [
        dim_size + ball_diameter for dim_size in get_shape(xs, ys, zs)
    ]
    volume = np.zeros(expanded_shape, dtype=np.uint32)

    x_min, y_min, z_min = xs.min(), ys.min(), zs.min()

    relative_xs = np.array((xs - x_min + ball_radius), dtype=np.int64)
    relative_ys = np.array((ys - y_min + ball_radius), dtype=np.int64)
    relative_zs = np.array((zs - z_min + ball_radius), dtype=np.int64)

    # OPTIMISE: vectorize
    for rel_x, rel_y, rel_z in zip(relative_xs, relative_ys, relative_zs):
        volume[rel_x, rel_y, rel_z] = np.iinfo(volume.dtype).max - 1
    return volume


def ball_filter_imgs(
    volume: np.ndarray,
    threshold_value: int,
    soma_centre_value: int,
    ball_xy_size: int = 3,
    ball_z_size: int = 3,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Apply ball filtering to a 3D volume and detect cell centres.

    Uses the `BallFilter` class to perform ball filtering on the volume
    and the `CellDetector` class to detect cell centres.

    Args:
        volume (np.ndarray): The 3D volume to be filtered.
        threshold_value (int): The threshold value for ball filtering.
        soma_centre_value (int): The value representing the soma centre.
        ball_xy_size (int, optional):
            The size of the ball filter in the XY plane. Defaults to 3.
        ball_z_size (int, optional):
            The size of the ball filter in the Z plane. Defaults to 3.

    Returns:
        Tuple[np.ndarray, np.ndarray]:
            A tuple containing the filtered volume and the cell centres.

    """
    # OPTIMISE: reuse ball filter instance

    good_tiles_mask = np.ones((1, 1, volume.shape[2]), dtype=bool)

    plane_width, plane_height = volume.shape[:2]

    bf = BallFilter(
        plane_width,
        plane_height,
        ball_xy_size,
        ball_z_size,
        overlap_fraction=0.8,
        tile_step_width=plane_width,
        tile_step_height=plane_height,
        threshold_value=threshold_value,
        soma_centre_value=soma_centre_value,
    )
    cell_detector = CellDetector(
        plane_width, plane_height, start_z=ball_z_size // 2
    )

    # FIXME: hard coded type
    ball_filtered_volume = np.zeros(volume.shape, dtype=np.uint32)
    previous_plane = None
    for z in range(volume.shape[2]):
        bf.append(volume[:, :, z].astype(np.uint32), good_tiles_mask[:, :, z])
        if bf.ready:
            bf.walk()
            middle_plane = bf.get_middle_plane()
            ball_filtered_volume[:, :, z] = middle_plane[:]
            # DEBUG: TEST: transpose
            previous_plane = cell_detector.process(
                middle_plane.copy(), previous_plane
            )
    return ball_filtered_volume, cell_detector.get_cell_centres()


def iterative_ball_filter(
    volume: np.ndarray, n_iter: int = 10
) -> Tuple[List[int], List[np.ndarray]]:
    """
    Apply iterative ball filtering to the given volume.
    The volume is eroded at each iteration, by subtracting 1 from the volume.

    Parameters:
        volume (np.ndarray): The input volume.
        n_iter (int): The number of iterations to perform. Default is 10.

    Returns:
        Tuple[List[int], List[np.ndarray]]: A tuple containing two lists:
            The structures found in each iteration.
            The cell centres found in each iteration.
    """
    ns = []
    centres = []

    threshold_value = np.iinfo(volume.dtype).max - 1
    soma_centre_value = np.iinfo(volume.dtype).max

    vol = volume.copy()  # TODO: check if required

    for i in range(n_iter):
        vol, cell_centres = ball_filter_imgs(
            vol, threshold_value, soma_centre_value
        )
        vol -= 1
        n_structures = len(cell_centres)
        ns.append(n_structures)
        centres.append(cell_centres)
        if n_structures == 0:
            break
    return ns, centres


def check_centre_in_cuboid(centre: np.ndarray, max_coords: np.ndarray) -> bool:
    """
    Checks whether a coordinate is in a cuboid
    :param centre: x,y,z coordinate
    :param max_coords: far corner of cuboid
    :return: True if within cuboid, otherwise False
    """
    relative_coords = centre
    if (relative_coords > max_coords).all():
        logger.info(
            'Relative coordinates "{}" exceed maximum volume '
            'dimension of "{}"'.format(relative_coords, max_coords)
        )
        return False
    else:
        return True


def split_cells(
    cell_points: np.ndarray, outlier_keep: bool = False
) -> np.ndarray:
    """
    Split the given cell points into individual cell centres.

    Args:
        cell_points (np.ndarray): Array of cell points with shape (N, 3),
            where N is the number of cell points and each point is represented
            by its x, y, and z coordinates.
        outlier_keep (bool, optional): Flag indicating whether to keep outliers
            during the splitting process. Defaults to False.

    Returns:
        np.ndarray: Array of absolute cell centres with shape (M, 3),
            where M is the number of individual cells and each centre is
            represented by its x, y, and z coordinates.
    """
    orig_centre = get_structure_centre(cell_points)

    xs = cell_points[:, 0]
    ys = cell_points[:, 1]
    zs = cell_points[:, 2]

    orig_corner = np.array(
        [
            orig_centre[0] - (orig_centre[0] - xs.min()),
            orig_centre[1] - (orig_centre[1] - ys.min()),
            orig_centre[2] - (orig_centre[2] - zs.min()),
        ]
    )

    relative_orig_centre = np.array(
        [
            orig_centre[0] - orig_corner[0],
            orig_centre[1] - orig_corner[1],
            orig_centre[2] - orig_corner[2],
        ]
    )

    original_bounding_cuboid_shape = get_shape(xs, ys, zs)

    ball_radius = 1
    vol = coords_to_volume(xs, ys, zs, ball_radius=ball_radius)

    # centres is a list of arrays of centres (1 array of centres per ball run)
    ns, centres = iterative_ball_filter(vol)
    ns.insert(0, 1)
    centres.insert(0, np.array([relative_orig_centre]))

    best_iteration = ns.index(max(ns))

    # TODO: put constraint on minimum centres distance ?
    relative_centres = centres[best_iteration]

    if not outlier_keep:
        # TODO: change to checking whether in original cluster shape
        original_max_coords = np.array(original_bounding_cuboid_shape)
        relative_centres = np.array(
            [
                x
                for x in relative_centres
                if check_centre_in_cuboid(x, original_max_coords)
            ]
        )

    absolute_centres = np.empty((len(relative_centres), 3))
    # FIXME: extract functionality
    absolute_centres[:, 0] = orig_corner[0] + relative_centres[:, 0]
    absolute_centres[:, 1] = orig_corner[1] + relative_centres[:, 1]
    absolute_centres[:, 2] = orig_corner[2] + relative_centres[:, 2]

    return absolute_centres



from typing import Tuple

import numpy as np


def make_sphere(
    ball_shape: Tuple[int, int, int],
    radius: float,
    position: Tuple[int, int, int],
) -> np.ndarray:
    """
    Return a boolean array, with array elements inside a sphere set
    to 1 and those outside set to 0.

    Parameters
    ----------
    ball_shape :
        Shape of the output array.
    radius :
        Radius of the sphere.
    position :
        Centre of the sphere.
    """

    half_sizes = (radius,) * 3

    # generate the grid for the support points
    # centered at the position indicated by position
    grid = [slice(-x0, dim - x0) for x0, dim in zip(position, ball_shape)]
    meshedgrid = np.ogrid[grid]
    # calculate the distance of all points from `position` center
    # scaled by the radius
    arr = np.zeros(ball_shape, dtype=float)
    for x_i, half_size in zip(meshedgrid, half_sizes):
        arr += np.abs(x_i / half_size) ** 2
    # the inner part of the sphere will have distance below 1
    return arr <= 1.0


def four_connected_kernel():
    return np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=bool)



from typing import List, Tuple

import napari
import numpy as np
import pandas as pd
from brainglobe_utils.cells.cells import Cell
from pkg_resources import resource_filename

brainglobe_logo = resource_filename(
    "cellfinder", "napari/images/brainglobe.png"
)


widget_header = """
<p>Efficient cell detection in large images.</p>
<p><a href="https://brainglobe.info" style="color:gray;">Website</a></p>
<p><a href="https://brainglobe.info/documentation/cellfinder/user-guide/napari-plugin/index.html" style="color:gray;">Documentation</a></p>
<p><a href="https://github.com/brainglobe/cellfinder" style="color:gray;">Source</a></p>
<p><a href="https://doi.org/10.1371/journal.pcbi.1009074" style="color:gray;">Citation</a></p>
<p><small>For help, hover the cursor over each parameter.</small>
"""  # noqa: E501


def html_label_widget(label: str, *, tag: str = "b") -> dict:
    """
    Create a HMTL label for use with magicgui.
    """
    return dict(
        widget_type="Label",
        label=f"<{tag}>{label}</{tag}>",
    )


header_label_widget = html_label_widget(
    f"""
<img src="{brainglobe_logo}"width="100">
<p>cellfinder</p>
""",
    tag="h1",
)


def add_layers(points: List[Cell], viewer: napari.Viewer) -> None:
    """
    Adds classified cell candidates as two separate point layers to the napari
    viewer.
    """
    detected, rejected = cells_to_array(points)

    viewer.add_points(
        rejected,
        name="Rejected",
        size=15,
        n_dimensional=True,
        opacity=0.6,
        symbol="ring",
        face_color="lightskyblue",
        visible=False,
        metadata=dict(point_type=Cell.UNKNOWN),
    )
    viewer.add_points(
        detected,
        name="Detected",
        size=15,
        n_dimensional=True,
        opacity=0.6,
        symbol="ring",
        face_color="lightgoldenrodyellow",
        metadata=dict(point_type=Cell.CELL),
    )


def cells_df_as_np(
    cells_df: pd.DataFrame,
    new_order: List[int] = [2, 1, 0],
    type_column: str = "type",
) -> np.ndarray:
    """
    Convert a dataframe to an array, dropping *type_column* and re-ordering
    the columns with *new_order*.
    """
    cells_df = cells_df.drop(columns=[type_column])
    cells = cells_df[cells_df.columns[new_order]]
    cells = cells.to_numpy()
    return cells


def cells_to_array(cells: List[Cell]) -> Tuple[np.ndarray, np.ndarray]:
    df = pd.DataFrame([c.to_dict() for c in cells])
    points = cells_df_as_np(df[df["type"] == Cell.CELL])
    rejected = cells_df_as_np(df[df["type"] == Cell.UNKNOWN])
    return points, rejected



from dataclasses import dataclass
from typing import Dict, Optional, TypeVar

import numba.typed
import numpy as np
import numpy.typing as npt
from numba import njit
from numba.core import types
from numba.experimental import jitclass
from numba.types import DictType


@dataclass
class Point:
    x: int
    y: int
    z: int


@njit
def get_non_zero_dtype_min(values: np.ndarray) -> int:
    """
    Get the minimum of non-zero entries in *values*.

    If all entries are zero, returns maximum storeable number
    in the values array.
    """
    min_val = np.iinfo(values.dtype).max
    for v in values:
        if v != 0 and v < min_val:
            min_val = v
    return min_val


T = TypeVar("T")


@njit
def traverse_dict(d: Dict[T, T], a: T) -> T:
    """
    Traverse d, until a is not present as a key.
    """
    if a in d:
        return traverse_dict(d, d[a])
    else:
        return a


@njit
def get_structure_centre(structure: np.ndarray) -> np.ndarray:
    """
    Get the pixel coordinates of the centre of a structure.

    Centre calculated as the mean of each pixel coordinate,
    rounded to the nearest integer.
    """
    # can't do np.mean(structure, axis=0)
    # because axis is not supported by numba
    return np.round(
        np.array(
            [
                np.mean(structure[:, 0]),
                np.mean(structure[:, 1]),
                np.mean(structure[:, 2]),
            ]
        )
    )


# Type declaration has to come outside of the class,
# see https://github.com/numba/numba/issues/8808
uint_2d_type = types.uint64[:, :]


spec = [
    ("z", types.uint64),
    ("next_structure_id", types.uint64),
    ("shape", types.UniTuple(types.int64, 2)),
    ("obsolete_ids", DictType(types.int64, types.int64)),
    ("coords_maps", DictType(types.uint64, uint_2d_type)),
]


@jitclass(spec=spec)
class CellDetector:
    """
    A class to detect connected structures within a series of
    stacked planes.

    Attributes
    ----------
    z :
        z-index of the plane currently being processed.
    next_structure_id :
        The next available structure ID that has yet to be used. IDs start
        counting up from 1.
    shape :
        Shape of the planes to be processed.
    obsolete_ids :
        Mapping from obsolete structure IDs to the structure ID they
        have been replaced with. This is used to keep track of structures
        that were initially disconnected, but later connected as the planes
        are scanned.
    coords_maps :
        Mapping from structure ID to the coordinates of pixels within that
        structure. Coordinates are stored in a 2D array, with the second
        axis indexing (x, y, z) coordinates.
    """

    def __init__(self, width: int, height: int, start_z: int):
        """
        Parameters
        ----------
        width, height
            Shape of the planes input to self.process()
        start_z:
            The z-coordinate of the first processed plane.
        """
        self.shape = width, height
        self.z = start_z
        self.next_structure_id = 1

        # Mapping from obsolete IDs to the IDs that they have been
        # made obsolete by
        self.obsolete_ids = numba.typed.Dict.empty(
            key_type=types.int64, value_type=types.int64
        )
        # Mapping from IDs to list of points in that structure
        self.coords_maps = numba.typed.Dict.empty(
            key_type=types.int64, value_type=uint_2d_type
        )

    def process(
        self, plane: np.ndarray, previous_plane: Optional[np.ndarray]
    ) -> np.ndarray:
        """
        Process a new plane.
        """
        if [e for e in plane.shape[:2]] != [e for e in self.shape]:
            raise ValueError("plane does not have correct shape")

        plane = self.connect_four(plane, previous_plane)
        self.z += 1
        return plane

    def connect_four(
        self, plane: np.ndarray, previous_plane: Optional[np.ndarray]
    ) -> np.ndarray:
        """
        Perform structure labelling.

        For all the pixels in the current plane, finds all structures touching
        this pixel using the four connected (plus shape) rule and also looks at
        the pixel at the same location in the previous plane. If structures are
        found, they are added to the structure manager and the pixel labeled
        accordingly.

        Returns
        -------
        plane :
            Plane with pixels either set to zero (no structure) or labelled
            with their structure ID.
        """
        SOMA_CENTRE_VALUE = np.iinfo(plane.dtype).max
        for y in range(plane.shape[1]):
            for x in range(plane.shape[0]):
                if plane[x, y] == SOMA_CENTRE_VALUE:
                    # Labels of structures below, left and behind
                    neighbour_ids = np.zeros(3, dtype=np.uint64)
                    # If in bounds look at neighbours
                    if x > 0:
                        neighbour_ids[0] = plane[x - 1, y]
                    if y > 0:
                        neighbour_ids[1] = plane[x, y - 1]
                    if previous_plane is not None:
                        neighbour_ids[2] = previous_plane[x, y]

                    if is_new_structure(neighbour_ids):
                        neighbour_ids[0] = self.next_structure_id
                        self.next_structure_id += 1
                    struct_id = self.add(x, y, self.z, neighbour_ids)
                else:
                    # reset so that grayscale value does not count as
                    # structure in next iterations
                    struct_id = 0

                plane[x, y] = struct_id

        return plane

    def get_cell_centres(self) -> np.ndarray:
        return self.structures_to_cells()

    def get_coords_dict(self) -> Dict:
        return self.coords_maps

    def add_point(self, sid: int, point: np.ndarray) -> None:
        """
        Add *point* to the structure with the given *sid*.
        """
        self.coords_maps[sid] = np.row_stack((self.coords_maps[sid], point))

    def add(
        self, x: int, y: int, z: int, neighbour_ids: npt.NDArray[np.uint64]
    ) -> int:
        """
        For the current coordinates takes all the neighbours and find the
        minimum structure including obsolete structures mapping to any of
        the neighbours recursively.

        Once the correct structure id is found, append a point with the
        current coordinates to the coordinates map entry for the correct
        structure. Hence each entry of the map will be a vector of all the
        pertaining points.
        """
        updated_id = self.sanitise_ids(neighbour_ids)
        if updated_id not in self.coords_maps:
            self.coords_maps[updated_id] = np.zeros(
                shape=(0, 3), dtype=np.uint64
            )
        self.merge_structures(updated_id, neighbour_ids)

        # Add point for that structure
        point = np.array([[x, y, z]], dtype=np.uint64)
        self.add_point(updated_id, point)
        return updated_id

    def sanitise_ids(self, neighbour_ids: npt.NDArray[np.uint64]) -> int:
        """
        Get the smallest ID of all the structures that are connected to IDs
        in `neighbour_ids`.

        For all the neighbour ids, walk up the chain of obsolescence (self.
        obsolete_ids) to reassign the corresponding most obsolete structure
        to the current neighbour.

        Has no side effects on this class.
        """
        for i, neighbour_id in enumerate(neighbour_ids):
            # walk up the chain of obsolescence
            neighbour_id = int(traverse_dict(self.obsolete_ids, neighbour_id))
            neighbour_ids[i] = neighbour_id

        # Get minimum of all non-obsolete IDs
        updated_id = get_non_zero_dtype_min(neighbour_ids)
        return int(updated_id)

    def merge_structures(
        self, updated_id: int, neighbour_ids: npt.NDArray[np.uint64]
    ) -> None:
        """
        For all the neighbours, reassign all the points of neighbour to
        updated_id. Then deletes the now obsolete entry from the points
        map and add that entry to the obsolete_ids.

        Updates:
        - self.coords_maps
        - self.obsolete_ids
        """
        for neighbour_id in np.unique(neighbour_ids):
            # minimise ID so if neighbour with higher ID, reassign its points
            # to current
            if neighbour_id > updated_id:
                self.add_point(updated_id, self.coords_maps[neighbour_id])
                self.coords_maps.pop(neighbour_id)
                self.obsolete_ids[neighbour_id] = updated_id

    def structures_to_cells(self) -> np.ndarray:
        cell_centres = np.empty((len(self.coords_maps.keys()), 3))
        for idx, structure in enumerate(self.coords_maps.values()):
            p = get_structure_centre(structure)
            cell_centres[idx] = p
        return cell_centres


@njit
def is_new_structure(neighbour_ids: np.ndarray) -> bool:
    for i in range(len(neighbour_ids)):
        if neighbour_ids[i] != 0:
            return False
    return True



import numpy as np
from numba import njit

from cellfinder.core.tools.array_operations import bin_mean_3d
from cellfinder.core.tools.geometry import make_sphere

DEBUG = False


class BallFilter:
    """
    A 3D ball filter.

    This runs a spherical kernel across the (x, y) dimensions
    of a *ball_z_size* stack of planes, and marks pixels in the middle
    plane of the stack that have a high enough intensity within the
    spherical kernel.
    """

    def __init__(
        self,
        plane_width: int,
        plane_height: int,
        ball_xy_size: int,
        ball_z_size: int,
        overlap_fraction: float,
        tile_step_width: int,
        tile_step_height: int,
        threshold_value: int,
        soma_centre_value: int,
    ):
        """
        Parameters
        ----------
        plane_width, plane_height :
            Width/height of the planes.
        ball_xy_size :
            Diameter of the spherical kernel in the x/y dimensions.
        ball_z_size :
            Diameter of the spherical kernel in the z dimension.
            Equal to the number of planes that stacked to filter
            the central plane of the stack.
        overlap_fraction :
            The fraction of pixels within the spherical kernel that
            have to be over *threshold_value* for a pixel to be marked
            as having a high intensity.
        tile_step_width, tile_step_height :
            Width/height of individual tiles in the mask generated by
            2D filtering.
        threshold_value :
            Value above which an individual pixel is considered to have
            a high intensity.
        soma_centre_value :
            Value used to mark pixels with a high enough intensity.
        """
        self.ball_xy_size = ball_xy_size
        self.ball_z_size = ball_z_size
        self.overlap_fraction = overlap_fraction
        self.tile_step_width = tile_step_width
        self.tile_step_height = tile_step_height

        self.THRESHOLD_VALUE = threshold_value
        self.SOMA_CENTRE_VALUE = soma_centre_value

        # Create a spherical kernel.
        #
        # This is done by:
        # 1. Generating a binary sphere at a resolution *upscale_factor* larger
        #    than desired.
        # 2. Downscaling the binary sphere to get a 'fuzzy' sphere at the
        #    original intended scale
        upscale_factor: int = 7
        upscaled_kernel_shape = (
            upscale_factor * ball_xy_size,
            upscale_factor * ball_xy_size,
            upscale_factor * ball_z_size,
        )
        upscaled_ball_centre_position = (
            np.floor(upscaled_kernel_shape[0] / 2),
            np.floor(upscaled_kernel_shape[1] / 2),
            np.floor(upscaled_kernel_shape[2] / 2),
        )
        upscaled_ball_radius = upscaled_kernel_shape[0] / 2.0
        sphere_kernel = make_sphere(
            upscaled_kernel_shape,
            upscaled_ball_radius,
            upscaled_ball_centre_position,
        )
        sphere_kernel = sphere_kernel.astype(np.float64)
        self.kernel = bin_mean_3d(
            sphere_kernel,
            bin_height=upscale_factor,
            bin_width=upscale_factor,
            bin_depth=upscale_factor,
        )

        assert (
            self.kernel.shape[2] == ball_z_size
        ), "Kernel z dimension should be {}, got {}".format(
            ball_z_size, self.kernel.shape[2]
        )

        self.overlap_threshold = np.sum(self.overlap_fraction * self.kernel)

        # Stores the current planes that are being filtered
        self.volume = np.empty(
            (plane_width, plane_height, ball_z_size), dtype=np.uint32
        )
        # Index of the middle plane in the volume
        self.middle_z_idx = int(np.floor(ball_z_size / 2))

        # TODO: lazy initialisation
        self.inside_brain_tiles = np.empty(
            (
                int(np.ceil(plane_width / tile_step_width)),
                int(np.ceil(plane_height / tile_step_height)),
                ball_z_size,
            ),
            dtype=bool,
        )
        # Stores the z-index in volume at which new planes are inserted when
        # append() is called
        self.__current_z = -1

    @property
    def ready(self) -> bool:
        """
        Return `True` if enough planes have been appended to run the filter.
        """
        return self.__current_z == self.ball_z_size - 1

    def append(self, plane: np.ndarray, mask: np.ndarray) -> None:
        """
        Add a new 2D plane to the filter.
        """
        if DEBUG:
            assert [e for e in plane.shape[:2]] == [
                e for e in self.volume.shape[:2]
            ], 'plane shape mismatch, expected "{}", got "{}"'.format(
                [e for e in self.volume.shape[:2]],
                [e for e in plane.shape[:2]],
            )
            assert [e for e in mask.shape[:2]] == [
                e for e in self.inside_brain_tiles.shape[:2]
            ], 'mask shape mismatch, expected"{}", got {}"'.format(
                [e for e in self.inside_brain_tiles.shape[:2]],
                [e for e in mask.shape[:2]],
            )
        if not self.ready:
            self.__current_z += 1
        else:
            # Shift everything down by one to make way for the new plane
            self.volume = np.roll(
                self.volume, -1, axis=2
            )  # WARNING: not in place
            self.inside_brain_tiles = np.roll(
                self.inside_brain_tiles, -1, axis=2
            )
        # Add the new plane to the top of volume and inside_brain_tiles
        self.volume[:, :, self.__current_z] = plane[:, :]
        self.inside_brain_tiles[:, :, self.__current_z] = mask[:, :]

    def get_middle_plane(self) -> np.ndarray:
        """
        Get the plane in the middle of self.volume.
        """
        z = self.middle_z_idx
        return np.array(self.volume[:, :, z], dtype=np.uint32)

    def walk(self) -> None:  # Highly optimised because most time critical
        ball_radius = self.ball_xy_size // 2
        # Get extents of image that are covered by tiles
        tile_mask_covered_img_width = (
            self.inside_brain_tiles.shape[0] * self.tile_step_width
        )
        tile_mask_covered_img_height = (
            self.inside_brain_tiles.shape[1] * self.tile_step_height
        )
        # Get maximum offsets for the ball
        max_width = tile_mask_covered_img_width - self.ball_xy_size
        max_height = tile_mask_covered_img_height - self.ball_xy_size

        _walk(
            max_height,
            max_width,
            self.tile_step_width,
            self.tile_step_height,
            self.inside_brain_tiles,
            self.volume,
            self.kernel,
            ball_radius,
            self.middle_z_idx,
            self.overlap_threshold,
            self.THRESHOLD_VALUE,
            self.SOMA_CENTRE_VALUE,
        )


@njit(cache=True)
def _cube_overlaps(
    cube: np.ndarray,
    overlap_threshold: float,
    THRESHOLD_VALUE: int,
    kernel: np.ndarray,
) -> bool:  # Highly optimised because most time critical
    """
    For each pixel in cube that is greater than THRESHOLD_VALUE, sum
    up the corresponding pixels in *kernel*. If the total is less than
    overlap_threshold, return False, otherwise return True.

    Halfway through scanning the z-planes, if the total overlap is
    less than 0.4 * overlap_threshold, this will return False early
    without scanning the second half of the z-planes.

    Parameters
    ----------
    cube :
        3D array.
    overlap_threshold :
        Threshold above which to return True.
    THRESHOLD_VALUE :
        Value above which a pixel is marked as being part of a cell.
    kernel :
        3D array, with the same shape as *cube*.
    """
    current_overlap_value = 0

    middle = np.floor(cube.shape[2] / 2) + 1
    halfway_overlap_thresh = (
        overlap_threshold * 0.4
    )  # FIXME: do not hard code value

    for z in range(cube.shape[2]):
        # TODO: OPTIMISE: step from middle to outer boundaries to check
        # more data first
        #
        # If halfway through the array, and the overlap value isn't more than
        # 0.4 * the overlap threshold, return
        if z == middle and current_overlap_value < halfway_overlap_thresh:
            return False  # DEBUG: optimisation attempt
        for y in range(cube.shape[1]):
            for x in range(cube.shape[0]):
                # includes self.SOMA_CENTRE_VALUE
                if cube[x, y, z] >= THRESHOLD_VALUE:
                    current_overlap_value += kernel[x, y, z]
    return current_overlap_value > overlap_threshold


@njit
def _is_tile_to_check(
    x: int,
    y: int,
    middle_z: int,
    tile_step_width: int,
    tile_step_height: int,
    inside_brain_tiles: np.ndarray,
) -> bool:  # Highly optimised because most time critical
    """
    Check if the tile containing pixel (x, y) is a tile that needs checking.
    """
    x_in_mask = x // tile_step_width  # TEST: test bounds (-1 range)
    y_in_mask = y // tile_step_height  # TEST: test bounds (-1 range)
    return inside_brain_tiles[x_in_mask, y_in_mask, middle_z]


@njit
def _walk(
    max_height: int,
    max_width: int,
    tile_step_width: int,
    tile_step_height: int,
    inside_brain_tiles: np.ndarray,
    volume: np.ndarray,
    kernel: np.ndarray,
    ball_radius: int,
    middle_z: int,
    overlap_threshold: float,
    THRESHOLD_VALUE: int,
    SOMA_CENTRE_VALUE: int,
) -> None:
    """
    Scan through *volume*, and mark pixels where there are enough surrounding
    pixels with high enough intensity.

    The surrounding area is defined by the *kernel*.

    Parameters
    ----------
    max_height, max_width :
        Maximum offsets for the ball filter.
    inside_brain_tiles :
        Array containing information on whether a tile is inside the brain
        or not. Tiles outside the brain are skipped.
    volume :
        3D array containing the plane-filtered data.
    kernel :
        3D array
    ball_radius :
        Radius of the ball in the xy plane.
    SOMA_CENTRE_VALUE :
        Value that is used to mark pixels in *volume*.

    Notes
    -----
    Warning: modifies volume in place!
    """
    for y in range(max_height):
        for x in range(max_width):
            ball_centre_x = x + ball_radius
            ball_centre_y = y + ball_radius
            if _is_tile_to_check(
                ball_centre_x,
                ball_centre_y,
                middle_z,
                tile_step_width,
                tile_step_height,
                inside_brain_tiles,
            ):
                cube = volume[
                    x : x + kernel.shape[0],
                    y : y + kernel.shape[1],
                    :,
                ]
                if _cube_overlaps(
                    cube,
                    overlap_threshold,
                    THRESHOLD_VALUE,
                    kernel,
                ):
                    volume[ball_centre_x, ball_centre_y, middle_z] = (
                        SOMA_CENTRE_VALUE
                    )



import numpy as np
from scipy.ndimage import gaussian_filter, laplace
from scipy.signal import medfilt2d


def enhance_peaks(
    img: np.ndarray, clipping_value: float, gaussian_sigma: float = 2.5
) -> np.ndarray:
    """
    Enhances the peaks (bright pixels) in an input image.

    Parameters:
    ----------
    img : np.ndarray
        Input image.
    clipping_value : float
        Maximum value for the enhanced image.
    gaussian_sigma : float, optional
        Standard deviation for the Gaussian filter. Default is 2.5.

    Returns:
    -------
    np.ndarray
        Enhanced image with peaks.

    Notes:
    ------
    The enhancement process includes the following steps:
    1. Applying a 2D median filter.
    2. Applying a Laplacian of Gaussian filter (LoG).
    3. Multiplying by -1 (bright spots respond negative in a LoG).
    4. Rescaling image values to range from 0 to clipping value.
    """
    type_in = img.dtype
    filtered_img = medfilt2d(img.astype(np.float64))
    filtered_img = gaussian_filter(filtered_img, gaussian_sigma)
    filtered_img = laplace(filtered_img)
    filtered_img *= -1

    filtered_img -= filtered_img.min()
    filtered_img /= filtered_img.max()

    # To leave room to label in the 3d detection.
    filtered_img *= clipping_value
    return filtered_img.astype(type_in)
